﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Threading;
using Net.Core.Communication;
using Net.Core.Communication.Serial.Sync;
using PH.Bootloader.RS485.DSP.IntelHex;
using PH.Bootloader.RS485.DSP.Procedure;
using PH.Bootloader.RS485.DSP.Protocol;

namespace PH.Bootloader.RS485.DSP
{
    /// <summary>封裝 DSP 啟動程序安裝之類別。</summary>
    public class DspBootloader : IBootloader
    {
        #region Fields
        private const ushort    COMM_TIMEOUT                = 220;  // 即為 Data Loading 指令與指令之間的間期
        private const ushort    CHECK_RETRY_COUNT           = 5;
        private const ushort    CHECK_COMMAND_INTERVAL      = 10;
        private const ushort    UPGRADE_RETRY_COUNT         = 10;
        private const ushort    UPGRADE_COMMAND_INTERVAL    = 10;
        public const ushort     WRITE_MEMORY_BLOCK_LENGTH   = 64;

        private const int       RESET_DEVICE_DELAY          = 500;   // 單位：毫秒

        // Communication
        private ISyncCommunicationPlus commPort;
        private bool            isComReady;
        // Boot procedure
        private IntelHexParser  intelHexParser;
        private long            startAddress;
        private Dictionary<long, MemoryBlockContent> memoryDictionary = null;
        private bool            isAbort;
        // Packet
        private PacketFactory   packetFactory;
        private ScriptType      currScriptType;
        private ScriptType      prevScriptType;
        private DateTime        upgradeTime;
        // Memory
        private byte[]          startPatternArray;
        private byte[]          stopPatternArray;
        private byte[]          startAddressArray;
        private byte[]          bankAddressArray;
        private byte[]          dataCountArray;
        private long            totalMemorySize;
        private long            sumMemorySize;
        // App
        private byte[]          psuResetCmdArray;
        private byte[]          checkCmdArray;
        #endregion

        #region Properties
        /// <summary>取得、設定名稱。</summary>
        public string           Name { get; set; }
        /// <summary>取得程序執行中旗標。</summary>
        public bool             IsProcedureRunning { get; private set; }
        /// <summary>設定、取得是否由 PIC 擔任代理燒錄的腳色 (透過 CAN Bus)。</summary>
        public bool             IsCanAgent { get; set; }
        #endregion

        #region Constructor
        /// <summary>初始化 ProcedureController 類別的新執行個體。</summary>
        public DspBootloader()
        {
            InitializeComponent();
        }

        /// <summary>初始化 ProcedureController 類別的新執行個體。</summary>
        /// <param name="syncComProvider"></param>
        public DspBootloader(ISyncCommunicationPlus syncSerialPort)
        {
            commPort = syncSerialPort;
            commPort.SetSyncTime(COMM_TIMEOUT);
            isComReady = commPort.IsOpen;

            InitializeComponent();
        }
        #endregion

        #region Methods : Public 
        /// <summary>終止所有處理程序。</summary>
        public void Abort()
        {
            isAbort = true;
        }

        /// <summary>指定序列埠通訊。</summary>
        /// <param name="comPort"></param>
        /// <returns></returns>
        public bool AssignComPort(ISyncCommunicationPlus syncSerialPort)
        {
            commPort = syncSerialPort;
            commPort.SetSyncTime(COMM_TIMEOUT);
            isComReady = commPort.IsOpen;
            return isComReady;
        }

        /// <summary>檢查裝置。</summary>
        /// <param name="device">目標裝置。</param>
        /// <returns></returns>
        public bool CheckDevice(DeviceInformation device)
        {
            memoryDictionary = null;

            if (!CheckSerialPort()) {
                return false;
            }

            if (currScriptType == ScriptType.Idle)
            {
                isAbort = false;
                upgradeTime = DateTime.Now;

                Thread thread = new Thread(() => CheckProcess(device));
                thread.IsBackground = true;
                thread.Name = "CheckDeviceProcess";
                thread.Start();

                return true;
            }
            else
            {
                CompleteRunWorkCallback(ProcedureType.CheckDevice, BootloadResult.HasOthersProcessRunning);
                return false;
            }
        }

        /// <summary>檢查 App Code 是否存在。</summary>
        /// <param name="device"></param>
        /// <returns></returns>
        public bool CheckAppCode(DeviceInformation device)
        {
            if (!CheckSerialPort()) {
                return false;
            }

            byte id = 0xFF;
            while (DeviceInformation.ParseBy485ID(id) == null && id != 0) {
                id--;
            }

            return !ImplementCheckDevice(id);
        }

        /// <summary>解析記憶體內容。</summary>
        /// <param name="device"></param>
        /// <param name="filePath"></param>
        public void ParseHexToMemory(DeviceInformation device, string filePath)
        {
            memoryDictionary = null;
            totalMemorySize = 0;
            MetaBlockContent metaBlockContent;
            IntelHexRecord[] hexRecords;

            ShowMessageCallback("\r\nParse image file...\r\n");

            IsProcedureRunning = true;
            if (File.Exists(filePath))
            {
                if (VerifyFileName(filePath))
                {
                    bool isHexReady = LoadHexFile(filePath, out hexRecords, out metaBlockContent);
                    if (isHexReady)
                    {
                        ParseMetaBlockCallback(metaBlockContent);
                        ParseHexRecordCallback(hexRecords);

                        startAddress = metaBlockContent.StartAddress;
                        ConvertToStartAddressArray(startAddress);

                        // 避免原始資料被改到!!
                        List<IntelHexRecord> tempRecords = new List<IntelHexRecord>();
                        foreach (IntelHexRecord record in hexRecords)
                        {
                            IntelHexRecord newRecord = new IntelHexRecord();
                            newRecord.Copy(record);
                            tempRecords.Add(newRecord);
                        }

                        if (RegulateHexRecord(tempRecords.ToArray(), startAddress, out memoryDictionary))
                        {
                            ConvertToDataCountArray(memoryDictionary.Count * WRITE_MEMORY_BLOCK_LENGTH);
                            ShowMessageCallback("Image file is ready!!\r\n");
                            CompleteRunWorkCallback(ProcedureType.HexToMemory, BootloadResult.MemoryAllocationReady);
                        }
                        else
                        {
                            ShowMessageCallback("           Err: Regulate hex record failed!!\r\n");
                            CompleteRunWorkCallback(ProcedureType.HexToMemory, BootloadResult.MemoryAllocationFailed);
                        }
                    }
                    else
                    {
                        ShowMessageCallback("           Err: Regulate hex record failed!!\r\n");
                        CompleteRunWorkCallback(ProcedureType.HexToMemory, BootloadResult.MemoryAllocationFailed);
                    }
                }
                else
                {
                    ShowMessageCallback("           Err: Image file is not match!!\r\n");
                    CompleteRunWorkCallback(ProcedureType.HexToMemory, BootloadResult.ImageFileNotMatch);
                }
            }
            else
            {
                ShowMessageCallback("           Err: Image file dose not exist!!\r\n");
                CompleteRunWorkCallback(ProcedureType.HexToMemory, BootloadResult.ImageFileNotExist);
            }

            IsProcedureRunning = false;
        }

        /// <summary>上載記憶體資訊。。</summary>
        /// <param name="device">指定的裝置。</param>
        /// <returns>程序是否啟動。</returns>
        public bool Upload(DeviceInformation device)
        {
            if (!CheckSerialPort()) {
                return false;
            }

            if (memoryDictionary != null)
            {
                if (currScriptType == ScriptType.Idle)
                {
                    isAbort = false;
                    upgradeTime = DateTime.Now;

                    Thread thread = new Thread(() => UploadImageProcedure(device, startAddress, memoryDictionary));
                    thread.IsBackground = true;
                    thread.Name = "UploadImageProcess_DSP";
                    thread.Start();
                    return true;
                }
                else
                {
                    CompleteRunWorkCallback(ProcedureType.Upload, BootloadResult.HasOthersProcessRunning);
                }
            }
            else
            {
                CompleteRunWorkCallback(ProcedureType.HexToMemory, BootloadResult.ImageFileNotExist);
            }

            return false;
        }
        #endregion

        #region Methods : Private (Check Device)
        private void CheckProcess(DeviceInformation device)
        {
            currScriptType = ScriptType.ProductID;
            prevScriptType = ScriptType.ProductID;

            bool done = false;
            bool isSuccess = false;
            string outcome;
            int scriptCount = 0;
            ReceivedPacketContent packetContent;
            Command cmd = Command.ProductID;

            ShowMessageCallback("\r\nCheck device...");
            IsProcedureRunning = true;
            while (true)
            {
                switch (currScriptType)
                {
                    case ScriptType.ProductID:
                        cmd = Command.Parse(currScriptType);
                        if (RequestFunction(device, cmd, null, 0, out packetContent)) {
                            currScriptType = ScriptType.SerialNumber;
                            outcome = string.Format("{0:X2}{1:X2}{2:X2}{3:X2}",
                                      packetContent.Payload[3], packetContent.Payload[2], packetContent.Payload[1], packetContent.Payload[0]);
                            ShowMessageCallback(outcome);
                        }
                        break;

                    case ScriptType.SerialNumber:
                        cmd = Command.Parse(currScriptType);
                        if (RequestFunction(device, cmd, null, 0, out packetContent)) {
                            currScriptType = ScriptType.ModelName;
                            outcome = string.Format("{0}{1}{2}{3}{4}{5}{6}{7}",
                                      Convert.ToChar(packetContent.Payload[1]), Convert.ToChar(packetContent.Payload[0]), Convert.ToChar(packetContent.Payload[3]), Convert.ToChar(packetContent.Payload[2]),
                                      Convert.ToChar(packetContent.Payload[5]), Convert.ToChar(packetContent.Payload[4]), Convert.ToChar(packetContent.Payload[7]), Convert.ToChar(packetContent.Payload[6]));
                            ShowMessageCallback(outcome.TrimEnd('\0'));
                        }
                        break;

                    case ScriptType.ModelName:
                        cmd = Command.Parse(currScriptType);
                        if (RequestFunction(device, cmd, null, 0, out packetContent)) {
                            currScriptType = ScriptType.HwVer;
                            outcome = string.Format("{0}{1}{2}{3}{4}{5}{6}{7}",
                                      Convert.ToChar(packetContent.Payload[9]), Convert.ToChar(packetContent.Payload[8]), Convert.ToChar(packetContent.Payload[11]), Convert.ToChar(packetContent.Payload[10]),
                                      Convert.ToChar(packetContent.Payload[13]), Convert.ToChar(packetContent.Payload[12]), Convert.ToChar(packetContent.Payload[15]), Convert.ToChar(packetContent.Payload[14]));
                            ShowMessageCallback(outcome);
                        }
                        break;

                    case ScriptType.HwVer:
                        cmd = Command.Parse(currScriptType);
                        if (RequestFunction(device, cmd, null, 0, out packetContent)) {
                            currScriptType = ScriptType.SwVer;
                            outcome = string.Format("{0:X2}.{1:X2}", packetContent.Payload[0], packetContent.Payload[1]);
                            ShowMessageCallback(outcome);
                        }
                        break;

                    case ScriptType.SwVer:
                        cmd = Command.Parse(currScriptType);
                        if (RequestFunction(device, cmd, null, 0, out packetContent)) {
                            currScriptType = ScriptType.Idle;
                            outcome = string.Format("PIC({0:X2}.{1:X2}) DSP({2:X2}.{3:X2})\r\n", packetContent.Payload[2], packetContent.Payload[3], packetContent.Payload[0], packetContent.Payload[1]);
                            ShowMessageCallback(outcome);

                            isSuccess = true;
                            done = true;
                        }
                        break;

                    case ScriptType.Idle:
                    default:
                        break;
                }

                if (!done)
                {
                    if (currScriptType == prevScriptType) {
                        ShowMessageCallback("Retry!!");

                        scriptCount++;
                        if (scriptCount >= CHECK_RETRY_COUNT) {
                            ShowMessageCallback("\r\n           Fail!!\r\n");
                            scriptCount = 0;
                            currScriptType += 1;
                            // 這個方法不好!!
                            if (currScriptType < ScriptType.CheckPassword1) {
                                prevScriptType = currScriptType;
                            } else {
                                done = true;
                            }
                        }
                    } else {
                        scriptCount = 0;
                        prevScriptType = currScriptType;
                    }

                    SpinWait.SpinUntil(() => false, CHECK_COMMAND_INTERVAL);
                }
                else
                {
                    break;
                }
                
                if (isAbort) {
                    isAbort = false;
                    break;
                }
            }

            IsProcedureRunning = false;
            currScriptType = ScriptType.Idle;
            CompleteRunWorkCallback(ProcedureType.CheckDevice, isSuccess? BootloadResult.DetectDevice : BootloadResult.DetectFailed);
        }
        #endregion

        #region Methods : Private (Parse Meta Block)
        /// <summary>確認檔案名稱。</summary>
        /// <param name="filePath"></param>
        /// <returns></returns>
        private bool VerifyFileName(string filePath)
        {
            if (!string.IsNullOrEmpty(filePath))
            {
                string strFileNameHead = Path.GetFileName(filePath).Substring(0, 3);
                return (string.Compare(strFileNameHead, DeviceType.PIC_CSU.HexKey.Substring(0, 3), StringComparison.OrdinalIgnoreCase) != 0) &&
                       (string.Compare(strFileNameHead, DeviceType.PIC_PSU.HexKey.Substring(0, 3), StringComparison.OrdinalIgnoreCase) != 0) &&
                       (string.Compare(strFileNameHead, DeviceType.PIC_SBox.HexKey.Substring(0, 3), StringComparison.OrdinalIgnoreCase) != 0);
            }

            return false;
        }

        /// <summary>載入 Hex 檔案。</summary>
        /// <param name="filePath"></param>
        /// <param name="startAddress"></param>
        /// <returns></returns>
        private bool LoadHexFile(string filePath, out IntelHexRecord[] intelHexRecords, out MetaBlockContent metaBlockContent)
        {
            intelHexRecords = null;
            metaBlockContent = null;

            intelHexParser.Parse(filePath, out intelHexRecords);
            if (intelHexRecords != null)
            {
                CompleteRunWorkCallback(ProcedureType.HexToMemory, BootloadResult.ImageFileReady);

                if (ParseMetaBlock(intelHexRecords, out metaBlockContent)) {
                    return true;
                } else {
                    CompleteRunWorkCallback(ProcedureType.HexToMemory, BootloadResult.MemoryAllocationFailed);
                    return false;
                }
            }
            else
            {
                CompleteRunWorkCallback(ProcedureType.HexToMemory, BootloadResult.ImageFileNotMatch);
                return false;
            }
        }

        /// <summary>解析出 Meta Block 功能。</summary>
        /// <param name="records"></param>
        /// <param name="outcome"></param>
        /// <returns></returns>
        private bool ParseMetaBlock(IntelHexRecord[] records, out MetaBlockContent outcome)
        {
            bool gotStartPattern = false;
            bool touchStopPattern = false;
            int metaBlockIndex = 0;
            byte[] metaBlockBuff = new byte[27]; // 只取中間資訊： 41-7-7 = 27
            outcome = null;

            // 2018/04/25 修改錯誤想法：『假設起始樣板一定存在於一個完整的 record 當中!!』
            // 答案是：並不一定存在於一個完整的 rescord 中....my bad
            try
            {
                int startPatternCount = 0;
                foreach (IntelHexRecord record in records)
                {
                    // 把 record 再打散成byte...@@
                    for (int i = 0; i< record.DataLength; i++)
                    {
                        // 2) 抽取出 metaBlock 資料
                        // 為何2)會在1)前面?因為要取從找到 startBlock 後的下一個data開始取metaBlock
                        if (gotStartPattern) {
                            metaBlockBuff[metaBlockIndex] = record.Data[i];
                            metaBlockIndex++;
                            if (metaBlockIndex >= 27) {
                                touchStopPattern = true;
                                break;
                            }
                        }

                        // 1) 先找出 Start Pattern
                        if (!gotStartPattern &&
                            record.Data[i] == startPatternArray[startPatternCount]) {
                            startPatternCount++;
                            if (startPatternCount >= startPatternArray.Length) {
                                gotStartPattern = true;
                            }
                        } else {
                            startPatternCount = 0;
                        }
                    }

                    if (touchStopPattern) {
                        break;
                    }
                }

                outcome = new MetaBlockContent();
                outcome.MetaBlockVersion = string.Format("V{0:X2}.{1:X2}", metaBlockBuff[1], metaBlockBuff[0]);
                outcome.SoftwareVersion = string.Format("V{0:X2}.{1:X2}", metaBlockBuff[10], metaBlockBuff[9]);
                outcome.StartAddressText = string.Format("0x{0:X2}{1:X2}{2:X2}{3:X2}", metaBlockBuff[16], metaBlockBuff[15], metaBlockBuff[14], metaBlockBuff[13]);
                outcome.StopAddressText = string.Format("0x{0:X2}{1:X2}{2:X2}{3:X2}", metaBlockBuff[20], metaBlockBuff[19], metaBlockBuff[18], metaBlockBuff[17]);
                outcome.StartAddress = Convert.ToInt64(string.Format("{0:X}{1:X2}{2:X2}{3:X2}", metaBlockBuff[16], metaBlockBuff[15], metaBlockBuff[14], metaBlockBuff[13]), 16);
                return true;
            }
            catch
            {
                return false;
            }
        }
  
        /// <summary>
        /// 正規畫資料內容。
        /// 1. 從位址資訊補正缺失的資料
        /// 2. 滿足資料封包大小等於 64bytes
        /// </summary>
        /// <param name="records"></param>
        /// <param name="startAddress"></param>
        /// <param name="memoryBank"></param>
        /// <returns></returns>
        private bool RegulateHexRecord(IntelHexRecord[] records, long startAddress, out Dictionary<long, MemoryBlockContent> memoryBank)
        {
            List<IntelHexRecord> fixedRecords = new List<IntelHexRecord>();
            List<byte> dataCollector = new List<byte>();
            long offsetAddress = 0;

            try
            {
                for (int i = 0; i < records.Length; i++)
                {
                    if (records[i].Type == IntelHexRecordType.ExtendedLinearAddressRecord)
                    {
                        offsetAddress = records[i].Data[0] << 8;
                        offsetAddress += records[i].Data[1];
                        offsetAddress = offsetAddress << 16;
                    }
                    else if (records[i].Type == IntelHexRecordType.DataRecord)
                    {
                        records[i].Address += offsetAddress;
                        // 大於起始位址
                        if (records[i].Address >= startAddress)
                        {
                            // 檢查前一筆資料是否有缺失資料
                            if (fixedRecords.Count > 0)
                            {
                                // 一個address存放一個word(=2 bytes)
                                IntelHexRecord prevRecord = fixedRecords[fixedRecords.Count - 1];
                                long expectedLen = (records[i].Address - prevRecord.Address) * 2;
                                int diff = (int)(expectedLen - prevRecord.DataLength);
                                if (diff > 0)
                                {
                                    int newLen = prevRecord.DataLength + diff;
                                    byte[] newArray = new byte[newLen];
                                    for (int j = 0; j < newArray.Length; j++)
                                    {
                                        newArray[j] = 0xFF;
                                    }
                                    Array.Copy(prevRecord.Data, 0, newArray, 0, prevRecord.DataLength);
                                    fixedRecords[fixedRecords.Count - 1].Data = newArray;
                                    fixedRecords[fixedRecords.Count - 1].DataLength = newLen;
                                }
                            }

                            fixedRecords.Add(records[i]);
                        }
                    }
                }

                dataCollector.Clear();
                foreach (IntelHexRecord record in fixedRecords)
                {
                    dataCollector.AddRange(record.Data);
                }

                // 滿足資料封包大小等於 64bytes
                while (dataCollector.Count % WRITE_MEMORY_BLOCK_LENGTH != 0)
                {
                    dataCollector.Add(0xFF);
                }

                // 以 64bytes 為一個Block重新分裝、定義位址
                memoryBank = null;
                GC.Collect();
                memoryBank = new Dictionary<long, MemoryBlockContent>();

                byte[] memDataArray = dataCollector.ToArray();
                for (int i = 0; i < dataCollector.Count; i += WRITE_MEMORY_BLOCK_LENGTH)
                {
                    MemoryBlockContent memBlock = new MemoryBlockContent(WRITE_MEMORY_BLOCK_LENGTH);
                    memBlock.Address = startAddress + i / 2;
                    Array.Copy(memDataArray, i, memBlock.DataArray, 0, WRITE_MEMORY_BLOCK_LENGTH);
                    memoryBank.Add(memBlock.Address, memBlock);
                }

                return true;
            }
            catch
            {
                memoryBank = null;
                return false;
            }
        }
        #endregion

        #region Methods : Private (Upload)
        /// <summary>程式更新程序。</summary>
        /// <param name="psuType"></param>
        /// <param name="startAddress"></param>
        /// <param name="memDict"></param>
        private void UploadImageProcedure(DeviceInformation psuType, long startAddress, Dictionary<long, MemoryBlockContent> memDict)
        {
            currScriptType = IsCanAgent ? ScriptType.AsignCanAgent : ScriptType.CheckPassword1;
            prevScriptType = IsCanAgent ? ScriptType.AsignCanAgent : ScriptType.CheckPassword1;

            bool done = false;
            bool isSuccess = false;
            int scriptCount = 0;
            int resendCount = -1;
            ReceivedPacketContent packetContent;
            Command cmd = Command.CheckPassword1;

            int dictIndex = 0;
            long writtingAddress = startAddress;
            int delayTime = 200;

            ShowMessageCallback("\r\nTransferring program memory...");

            IsProcedureRunning = true;
            while (true)
            {
                switch (currScriptType)
                {
                    case ScriptType.AsignCanAgent:
                        cmd = Command.Parse(currScriptType);
                        if (RequestFunction(psuType, cmd, startPatternArray, startPatternArray.Length, out packetContent)) {
                            currScriptType = ScriptType.CheckPassword1;
                            ShowMessageCallback("Running...");
                        }
                        break;

                    case ScriptType.CheckPassword1:
                        cmd = Command.Parse(currScriptType);
                        dictIndex = 0;
                        writtingAddress = startAddress;
                        if (RequestFunction(psuType, cmd, startPatternArray, startPatternArray.Length, out packetContent)) {
                            currScriptType = ScriptType.CheckPassword2;
                            ShowMessageCallback("Pass!!");
                        }
                        break;

                    case ScriptType.CheckPassword2:
                        cmd = Command.Parse(currScriptType);
                        if (RequestFunction(psuType, cmd, stopPatternArray, stopPatternArray.Length, out packetContent)) {
                            currScriptType = ScriptType.Investigate;
                            ShowMessageCallback("Pass!!");
                        }
                        break;

                    case ScriptType.Investigate:
                        cmd = Command.Parse(currScriptType);
                        if (RequestFunction(psuType, cmd, null, 0, out packetContent)) {
                            currScriptType = ScriptType.SelectBank;
                            ShowMessageCallback("OK!!");
                        }
                        break;

                    case ScriptType.SelectBank:
                        cmd = Command.Parse(currScriptType);
                        if (RequestFunction(psuType, cmd, bankAddressArray, bankAddressArray.Length, out packetContent)) {
                            if (CompareArray(packetContent.Payload.ToArray(), bankAddressArray, bankAddressArray.Length)) {
                                currScriptType = ScriptType.DataLoading;
                                sumMemorySize = 0;
                                delayTime = UPGRADE_COMMAND_INTERVAL;
                                ShowMessageCallback("OK!!");
                                ShowMessageCallback(string.Format("\r\n           {0}...", Command.DataLoading.Text));
                            }
                        }
                        break;

                    case ScriptType.DataLoading:
                        cmd = Command.Parse(currScriptType);
                        MemoryBlockContent memBlock;
                        if (!SerachMemoryBlock(memDict, writtingAddress, ref dictIndex, out memBlock)) {
                            if (ProcessNextddress(memDict, ref dictIndex, ref writtingAddress)) {
                                currScriptType = ScriptType.CloseBank;
                                delayTime = 500;
                                break;
                            }
                            SerachMemoryBlock(memDict, writtingAddress, ref dictIndex, out memBlock);
                        }

                        // 記憶體資料寫入，PIC 不會有回應，除非是重新要資料時
                        byte[] memBlkArray = memBlock.Merge();
                        if (RequestFunction(psuType, cmd, Command.ReplayDataLoading, memBlkArray, memBlkArray.Length, out packetContent)) {
                            writtingAddress = packetContent.Payload[0] + (packetContent.Payload[1] << 8) + (packetContent.Payload[2] << 16) + (packetContent.Payload[3] << 24);
                            resendCount++;
                            if (resendCount > 0) {
                                ShowMessageCallback(string.Format(" Resend#{0}... ", resendCount));
                            }
                            Thread.CurrentThread.Join(600);
                        } else {
                            // 提取位置，檢查是否到資料尾
                            if (ProcessNextddress(memDict, ref dictIndex, ref writtingAddress)) {
                                currScriptType = ScriptType.CloseBank;
                                ShowMessageCallback("OK!!");
                                delayTime = 500;
                                isSuccess = true;
                            }
                            
                            float percentage = (float)dictIndex / memDict.Count;
                            sumMemorySize = dictIndex * WRITE_MEMORY_BLOCK_LENGTH;
                            ChangeProcedureProgressCallback(ProcedureType.Upload, (int)(percentage*100.0F));
                            ChangeUploadInformationCallback(sumMemorySize, totalMemorySize);
                        }
                        break;

                    case ScriptType.CloseBank:
                        cmd = Command.Parse(currScriptType);
                        if (RequestFunction(psuType, cmd, dataCountArray, dataCountArray.Length, out packetContent)) {
                            currScriptType = ScriptType.UpgradeDone;
                            ShowMessageCallback("OK!!");
                        }
                        break;

                    case ScriptType.UpgradeDone:
                        cmd = Command.Parse(currScriptType);
                        if (RequestFunction(psuType, cmd, new byte[] { 0x00 }, 1, out packetContent)) {
                            if (IsCanAgent) {
                                ShowMessageCallback("OK!!");
                                currScriptType = ScriptType.CanAgentSummary;
                            } else {
                                ShowMessageCallback("OK!!");
                                ShowMessageCallback("\r\nTransferred program memory successfully ★★★★★\r\n");

                                currScriptType = ScriptType.Idle;
                                done = true;
                            }
                        }
                        break;

                    case ScriptType.WaitPowerRestore:
                        delayTime = 200;
                        cmd = Command.CheckPassword1;
                        ShowMessageCallback("\r\n           Wait power restore...\r\n");
                        if (RequestFunction(psuType, cmd, startPatternArray, startPatternArray.Length, out packetContent)) {
                            ShowMessageCallback("Pass!!");
                            dictIndex = 0;
                            writtingAddress = startAddress;
                            currScriptType = ScriptType.CheckPassword2;
                        }
                        break;

                    case ScriptType.CanAgentSummary:
                        cmd = Command.Parse(currScriptType);
                        if (RequestFunction(psuType, cmd, dataCountArray, dataCountArray.Length, out packetContent)) {
                            ParseCanAgentSummary(cmd, packetContent.Payload.ToArray());
                            ShowMessageCallback("\r\nTransferred program memory successfully ★★★★★\r\n");
                            currScriptType = ScriptType.Idle;
                            done = true;
                        }
                        break;

                    case ScriptType.Idle:
                    default:
                        break;
                }

                if (!done)
                {
                    if (currScriptType == prevScriptType) {
                        // Wait Power Restore 沒有重傳次數的限制，只有時間限制
                        if (currScriptType == ScriptType.WaitPowerRestore) {
                            if ((DateTime.Now - upgradeTime).TotalMinutes >= 1) {
                                ShowMessageCallback("\r\n           Retry more than 1 mins,\nupgrade stop!!\r\n");
                                break;
                            }
                        } else {
                            // DataLoading 重複是正常的，所以不支持重傳
                            if (currScriptType != ScriptType.DataLoading) {
                                ShowMessageCallback("Retry...");
                                scriptCount++;
                                if (scriptCount >= UPGRADE_RETRY_COUNT) {
                                    ShowMessageCallback("\r\n           Fail!!\r\n");
                                    currScriptType = currScriptType == ScriptType.CloseBank ? ScriptType.WaitPowerRestore : ScriptType.CloseBank;
                                    scriptCount = 0;
                                    prevScriptType = currScriptType;
                                }
                            }
                        }
                    } else {
                        scriptCount = 0;
                        prevScriptType = currScriptType;
                    }

                    Thread.CurrentThread.Join(delayTime);
                }
                else
                {
                    break;
                }

                if (isAbort) {
                    break;
                }
            }

            if (isAbort) {
                isAbort = false;
                ShowMessageCallback("\r\nTransferring program memory has been aborted!!\r\n");
            }

            IsProcedureRunning = false;
            currScriptType = ScriptType.Idle;
            ChangeProcedureProgressCallback(ProcedureType.Upload, 0);
            ChangeUploadInformationCallback(0, 0);
            CompleteRunWorkCallback(ProcedureType.Upload, isSuccess ? BootloadResult.Uploaded : BootloadResult.UploadFailed);
        }

        /// <summary>根據起始位址，搜尋記憶體區塊物件、並回傳該物件在字典容器內的位址。</summary>
        /// <param name="memDict"></param>
        /// <param name="address">起始位址。</param>
        /// <param name="dictIndex">字典容器中對應的指標。</param>
        /// <param name="memBlock">記憶體區塊物件。</param>
        /// <returns></returns>
        private bool SerachMemoryBlock(Dictionary<long, MemoryBlockContent> memDict, long address, ref int dictIndex, out MemoryBlockContent memBlock)
        {
            if (memDict.ContainsKey(address))
            {
                memBlock = memDict[address];
                dictIndex = memDict.Keys.ToList().IndexOf(address);
                return true;
            }
            else
            {
                memBlock = null;
                return false;
            }
        }

        /// <summary>自動位移至下一個起始位址，回傳起始位址及字典容器中對應的指標。</summary>
        /// <param name="memDict"></param>
        /// <param name="dictIndex">字典容器中對應的指標。</param>
        /// <param name="address">起始位址。</param>
        /// <returns>是否已經取完所有容器內資料。</returns>
        private bool ProcessNextddress(Dictionary<long, MemoryBlockContent> memDict, ref int dictIndex, ref long address)
        {
            dictIndex++;
            if (dictIndex < memDict.Count)
            {
                address = memDict.ToList()[dictIndex].Value.Address;
                return false;
            }
            else
            {
                return true;
            }
        }

        private void ParseCanAgentSummary(Command cmd, byte[] dataArray)
        {
            ShowMessageCallback(string.Format("Total client # = {0}", dataArray[0]));

            string devCollection = " (";
            for (int i = 1; i < dataArray.Length; i++)
            {
                if (dataArray[i] != 0x00) {
                    devCollection += string.Format("{0:X2} ", dataArray[i]);
                }
            }

            ShowMessageCallback(devCollection.TrimEnd(' ') + ")");
        }
        #endregion

        #region Methods : Private
        /// <summary>初始化物件。</summary>
        private void InitializeComponent()
        {
            startAddress = 0;
            intelHexParser = new IntelHexParser();

            packetFactory = PacketFactory.GetInstance();
            currScriptType = ScriptType.Idle;
            prevScriptType = ScriptType.Idle;

            startPatternArray = new byte[7] { 0x7C, 0x0F, 0x4C, 0xF0, 0x21, 0x4A, 0x55 };
            stopPatternArray = new byte[7] { 0x96, 0x5A, 0x8F, 0x30, 0xF0, 0x63, 0xE0 };
            startAddressArray = new byte[4];
            bankAddressArray = new byte[5];
            dataCountArray = new byte[5];

            psuResetCmdArray = new byte[10] { 0xF0, 0x10, 0xFE, 0x05, 0x01, 0x8A, 0xFF, 0x00, 0x00, 0x00 };
            checkCmdArray = new byte[1] { 0xC1 };
        }

        /// <summary>檢查串列埠是否備妥。</summary>
        /// <returns></returns>
        private bool CheckSerialPort()
        {
            if (!isComReady)
            {
                CompleteRunWorkCallback(ProcedureType.CheckCommunication, BootloadResult.SerialPortNotReady);
                ShowMessageCallback("           Err: Serial port is not ready yet!!\r\n");
            }
            else
            {
                CompleteRunWorkCallback(ProcedureType.CheckCommunication, BootloadResult.SerialPortReady);
            }

            return isComReady;
        }

        /// <summary>實作檢查裝置。</summary>
        /// <param name="devType"></param>
        /// <param name="devID"></param>
        /// <param name="isShowMsg"></param>
        /// <returns></returns>
        private bool ImplementCheckDevice(byte devID)
        {
            ResetDevice(devID);
            return SendBootKey();
        }

        /// <summary>傳送 Boot 代碼。</summary>
        /// <returns></returns>
        private bool SendBootKey()
        {
            byte[] outComePtr;
            return commPort.SendExpectedLengthReceive(checkCmdArray, 1, 2, out outComePtr);
        }

        /// <summary>重置裝置。</summary>
        /// <param name="devID"></param>
        private void ResetDevice(byte devID)
        {
            byte[] outComePtr;
            int crc16;

            psuResetCmdArray[2] = devID;
            crc16 = CRC16.Calculate(psuResetCmdArray, 8);
            psuResetCmdArray[8] = (byte)((crc16 & 0x00FF));
            psuResetCmdArray[9] = (byte)((crc16 & 0xFF00) >> 8);

            commPort.SendExpectedLengthReceive(psuResetCmdArray, 10, 10, out outComePtr, true);

            Thread.CurrentThread.Join(RESET_DEVICE_DELAY);
        }

        private void ShowRequest(ScriptType type)
        {
            if (type == ScriptType.Idle || 
                type == ScriptType.WaitPowerRestore ||
                type == ScriptType.DataLoading)
            {
                return;
            }

            Command cmd = Command.Parse(type);
            ShowMessageCallback(string.Format("\r\n           {0}: ", cmd.Text));
        }

        private void ConvertToStartAddressArray(long startAddress)
        {
            startAddressArray[0] = (byte)startAddress;
            startAddressArray[1] = (byte)(startAddress >> 8);
            startAddressArray[2] = (byte)(startAddress >> 16);
            startAddressArray[3] = (byte)(startAddress >> 24);
            bankAddressArray[0] = 0;
            bankAddressArray[1] = startAddressArray[0];
            bankAddressArray[2] = startAddressArray[1];
            bankAddressArray[3] = startAddressArray[2];
            bankAddressArray[4] = startAddressArray[3];
        }

        private void ConvertToDataCountArray(long dataCount)
        {
            dataCountArray[0] = 0;
            dataCountArray[1] = (byte)dataCount;
            dataCountArray[2] = (byte)(dataCount >> 8);
            dataCountArray[3] = (byte)(dataCount >> 16);
            dataCountArray[4] = (byte)(dataCount >> 24);

            totalMemorySize = dataCount;
        }

        private bool RequestFunction(DeviceInformation device, Command cmd, byte[] dataArray, int length, out ReceivedPacketContent packetContent)
        {
            return RequestFunction(device, cmd, cmd, dataArray, length, out packetContent);
        }

        private bool RequestFunction(DeviceInformation device, Command cmd, Command replyCmd, byte[] dataArray, int length, out ReceivedPacketContent packetContent)
        {
            ShowRequest(cmd.ScriptType);

            packetContent = null;
            byte[] packetArray = packetFactory.Create(device, cmd, dataArray, length);
            byte[] receivedBuff;

            // Expand for easy check
            bool check1 = commPort.SendExpectedLengthReceive(packetArray, packetArray.Length, cmd.ResponsePacketLength, out receivedBuff);
            if (!check1) { return false; }
            bool check2 = packetFactory.Parse(receivedBuff, receivedBuff.Length, out packetContent);
            if (!check2) { return false; }
            bool check3 = packetFactory.ValidateResponsePacket(device, cmd, replyCmd, packetContent);
            return (check1 && check2 && check3);
        }

        private bool CompareArray(byte[] array1, byte[] array2, int len)
        {
            if (array1.Length < len || array2.Length < len) { return false; }

            for (int i = 0; i< len; i++) {
                if (array1[i] != array2[i]) {
                    return false;
                }
            }
            return true;
        }
        #endregion

        #region Methods : Callback
        private void ParseHexCallback(string[] records)
        {
            HexParsed?.Invoke(this, records);
        }

        private void ShowMessageCallback(string message)
        {
            MessageShown?.Invoke(this, message);
        }

        private void CompleteRunWorkCallback(ProcedureType procType, BootloadResult result)
        {
            ProcedureCompleted?.Invoke(this, procType, result);
        }

        private void ChangeProcedureProgressCallback(ProcedureType procType, int progressPercentage)
        {
            if (progressPercentage < 0){
                progressPercentage = 0;
            } else if (progressPercentage > 100) {
                progressPercentage = 100;
            }

            ProcedureProgressChanged?.Invoke(this, procType, progressPercentage);
        }

        private void ChangeUploadInformationCallback(long current, long total)
        {
            UploadInformationChanged?.Invoke(this, current, total);
        }

        private void ParseMetaBlockCallback(MetaBlockContent metaBlockContent)
        {
            MetaBlockParsed?.Invoke(this, metaBlockContent);
        }

        private void ParseHexRecordCallback(IntelHexRecord[] hexRecords)
        {
            HexRecordParsed?.Invoke(this, hexRecords);
        }
        #endregion

        #region Events
        public delegate void MetaBlockParsedEventHandler(object sender, MetaBlockContent metaBlockContent);
        public delegate void IntelHexRecordParsedEventHandler(object sender, IntelHexRecord[] hexRecords);

        /// <summary>發生在解析出 Hex 內容後。</summary>
        public event HexParsedEventHandler                  HexParsed;
        /// <summary>發生在更新程序信息顯示時。</summary>
        public event MessageShownEventHandler               MessageShown;
        /// <summary>發生在程序執行完成時。</summary>
        public event ProcedureCompletedEventHandler         ProcedureCompleted;
        /// <summary>發生在程序進度改變時。</summary>
        public event ProcedureProgressChangedEventHandler   ProcedureProgressChanged;
        /// <summary>發生在資料更新改變時。</summary>
        public event UploadInformationChangedEventHandler   UploadInformationChanged;

        public event MetaBlockParsedEventHandler            MetaBlockParsed;
        public event IntelHexRecordParsedEventHandler       HexRecordParsed;
        #endregion
    }
}
