﻿using System;
using System.IO;
using System.Linq;

namespace PH.Bootloader.RS485.DSP.IntelHex
{
    /// <summary>表示 Intel Hex 檔案解析處理類別。</summary>
    public class IntelHexParser
    {
        #region Fields
        private const int START_CODE_OFFSET     = 0;
        private const int START_CODE_LENGTH     = 1;
        private const int BYTE_COUNT_OFFSET     = START_CODE_OFFSET + START_CODE_LENGTH;
        private const int BYTE_COUNT_LENGTH     = 2;
        private const int ADDRESS_OFFSET        = BYTE_COUNT_OFFSET + BYTE_COUNT_LENGTH;
        private const int ADDRESS_LENGTH        = 4;
        private const int RECORD_TYPE_OFFSET    = ADDRESS_OFFSET + ADDRESS_LENGTH;
        private const int RECORD_TYPE_LENGTH    = 2;
        private const int DATA_OFFSET           = RECORD_TYPE_OFFSET + RECORD_TYPE_LENGTH;
        private const int CHECKSUM_LENGTH       = 2;

        private int MinimumLineLength           = START_CODE_OFFSET + BYTE_COUNT_LENGTH + ADDRESS_LENGTH + RECORD_TYPE_LENGTH + CHECKSUM_LENGTH;
        #endregion

        #region Methods-Public
        /// <summary>解析字串。</summary>
        /// <param name="filePth"></param>
        /// <returns></returns>
        public byte[] Parse(string filePth)
        {
            IntelHexRecord[] records;
            return Parse(filePth, out records);
        }

        /// <summary>解析字串。</summary>
        /// <param name="filePth"></param>
        /// <param name="records"></param>
        /// <returns></returns>
        public byte[] Parse(string filePth, out IntelHexRecord[] records)
        {
            string source;

            try
            {
                using (StreamReader streamReader = new StreamReader(filePth)) {
                    source = streamReader.ReadToEnd();
                }
            }
            catch
            {
                records = null;
                return null;
            }

            int finalDataSize = ParseText(source, out records);
            if (finalDataSize > 0)
            {
                byte[] outcome = new byte[finalDataSize];
                int outcomeIndex = 0;
                for (int i = 0; i < records.Length; i++)
                {
                    IntelHexRecord record = records[i];
                    Array.Copy(record.Data, 0, outcome, outcomeIndex, record.DataLength);
                    outcomeIndex += record.DataLength;
                }

                return outcome;
            }

            throw new IntelHexParserException(IntelHexParserExceptionKind.UNKNOWN);
        }
        #endregion

        #region Methods-Private
        /// <summary>解析字串內容。</summary>
        /// <param name="source"></param>
        /// <param name="records"></param>
        /// <returns>資料大小。若等於0則表示解析失敗。</returns>
        private int ParseText(string source, out IntelHexRecord[] records)
        {
            string[] lines = source.Split(Environment.NewLine.ToCharArray());
            lines = lines.Where(line => line.Length > 0).ToArray();

            records = new IntelHexRecord[lines.Length];
            int recordCount = 0;
            int finalDataSize = 0;

            foreach (string line in lines)
            {
                if (line.Length < MinimumLineLength) {
                    throw new IntelHexParserException(IntelHexParserExceptionKind.INVALID_LINE);
                }

                if (line[0] != ':') {
                    throw new IntelHexParserException(IntelHexParserExceptionKind.MISSING_START_CODE);
                }

                IntelHexRecord record = new IntelHexRecord();
                record.DataLength = Convert.ToInt16(line.Substring(BYTE_COUNT_OFFSET, BYTE_COUNT_LENGTH), 16);
                record.Address = Convert.ToUInt16(line.Substring(ADDRESS_OFFSET, ADDRESS_LENGTH), 16);
                record.Type = (IntelHexRecordType)Convert.ToByte(line.Substring(RECORD_TYPE_OFFSET, RECORD_TYPE_LENGTH), 16);

                if (record.Type > IntelHexRecordType.StartLinearAddressRecord) {
                    throw new IntelHexParserException(IntelHexParserExceptionKind.UNSUPPORTED_RECORD_TYPE);
                }

                record.Data = new byte[record.DataLength];

                for (int i = 0; i < record.DataLength; i++) {
                    record.Data[i] = Convert.ToByte(line.Substring(DATA_OFFSET + 2 * i, 2), 16);
                }

                if (!IsChecksumValid(line, record)) {
                    throw new IntelHexParserException(IntelHexParserExceptionKind.INVALID_CHECKSUM);
                }

                records[recordCount++] = record;
                finalDataSize += record.DataLength;
            }

            return finalDataSize;
        }

        /// <summary>判斷校驗值是否正確。</summary>
        /// <param name="line"></param>
        /// <param name="record"></param>
        /// <returns></returns>
        private bool IsChecksumValid(string line, IntelHexRecord record)
        {
            byte current = Convert.ToByte(line.Substring(DATA_OFFSET + record.DataLength * 2, CHECKSUM_LENGTH), 16);
            return record.Checksum == current;
        }
        #endregion
    }
}
