﻿using System.Collections.Generic;
using PH.Bootloader.RS485.DSP.Procedure;

namespace PH.Bootloader.RS485.DSP.Protocol
{
    /// <summary>封裝所有指令型態之類別。</summary>
    public class Command 
    {
        /// <summary>DSP 要求指令重傳時的位移。</summary>
        private static byte             REPLY_OFFSET = 0x80;
        /// <summary>指令集。</summary>
        private static List<Command>    cmdGroup = new List<Command>();

        /// <summary>產品編號。</summary>
        public static Command   ProductID = new Command(0x00, ScriptType.ProductID, "Product ID", 7, 11);
        /// <summary>序號。</summary>
        public static Command   SerialNumber = new Command(0x01, ScriptType.SerialNumber, "Serial Number", 7, 15);
        /// <summary>型號。</summary>
        public static Command   ModelName = new Command(0x02, ScriptType.ModelName, "Model Name", 7, 23);
        /// <summary>硬體版本。</summary>
        public static Command   HwVer = new Command(0x03, ScriptType.HwVer, "Hardware Version", 7, 9);
        /// <summary>軟體版本。</summary>
        public static Command   SwVer = new Command(0x04, ScriptType.SwVer, "Software Version", 7, 11);
        /// <summary>密碼1。</summary>
        public static Command   CheckPassword1 = new Command(0x05, ScriptType.CheckPassword1, "Password1 Check", 14, 7);
        /// <summary>密碼2。</summary>
        public static Command   CheckPassword2 = new Command(0x06, ScriptType.CheckPassword2, "Password2 Check", 14, 7);
        /// <summary>設定 PIC 做為 CAN Master，進行燒錄廣播。(瑱賢新增)</summary>
        public static Command   AsignCanAgent = new Command(0x07, ScriptType.AsignCanAgent, "CAN Agent", 7, 7);
        /// <summary>取得 CAN Master 燒錄結果。(瑱賢新增)</summary>
        public static Command   CanAgentSummary = new Command(0x08, ScriptType.CanAgentSummary, "CAN Summary", 7, 40);
        /// <summary>資料上傳。</summary>
        public static Command   DataLoading = new Command(0x10, ScriptType.DataLoading, "Data Loading", 11 + DspBootloader.WRITE_MEMORY_BLOCK_LENGTH, 11);
        /// <summary>BankSelect。</summary>
        public static Command   SelectBank = new Command(0x12, ScriptType.SelectBank, "Bank Select", 11, 12);
        /// <summary>BankClose。</summary>
        public static Command   CloseBank = new Command(0x13, ScriptType.CloseBank, "Bank Close", 12, 10);
        /// <summary>更新完成。</summary>
        public static Command   UpgradeDone = new Command(0x14, ScriptType.UpgradeDone, "Upgrade Done", 8, 8);
        /// <summary>檢查。</summary>
        public static Command   Investigate = new Command(0x15, ScriptType.Investigate, "Investigate", 7, 9);
        /// <summary>資料上傳重傳回應(原本不在指令列內，為了Parsing方便)。</summary>
        public static Command   ReplayDataLoading = new Command((byte)(DataLoading.Code + REPLY_OFFSET), ScriptType.DataLoading, "Reply Data Loading", 0, 11);

        /// <summary>取得指令編碼。</summary>
        public byte             Code { get; internal set; }
        /// <summary>取得腳本型態。</summary>
        public ScriptType       ScriptType { get; internal set; }
        /// <summary>文字敘述。</summary>
        public string           Text { get; internal set; }
        /// <summary>取得傳送指令之封包總長度。</summary>
        public ushort           RequestPacketLength { get; internal set; }
        /// <summary>取得收到的回復封包總長度。</summary>
        public ushort           ResponsePacketLength { get; internal set; }

        /// <summary>初始化 Command 類別的新執行個體。</summary>
        /// <param name="code"></param>
        /// <param name="scriptType"></param>
        /// <param name="text"></param>
        /// <param name="requestLen"></param>
        /// <param name="responseLen"></param>
        private Command(byte code, ScriptType scriptType, string text, ushort requestLen, ushort responseLen)
        {
            Code = code;
            ScriptType = scriptType;
            Text = text;
            RequestPacketLength = requestLen;
            ResponsePacketLength = responseLen;
        }

        /// <summary>比較兩物件是否相同。</summary>
        /// <param name="cmd"></param>
        /// <returns></returns>
        public bool Equals(Command cmd)
        {
            return Code == cmd.Code;
        }

        /// <summary>取得指令列表。</summary>
        /// <returns></returns>
        public static Command[] GetList()
        {
            if (cmdGroup.Count <= 0)
            {
                cmdGroup.Add(ProductID);
                cmdGroup.Add(SerialNumber);
                cmdGroup.Add(ModelName);
                cmdGroup.Add(HwVer);
                cmdGroup.Add(SwVer);
                cmdGroup.Add(CheckPassword1);
                cmdGroup.Add(CheckPassword2);
                cmdGroup.Add(AsignCanAgent); 
                cmdGroup.Add(CanAgentSummary);
                cmdGroup.Add(DataLoading);
                cmdGroup.Add(SelectBank);
                cmdGroup.Add(CloseBank);
                cmdGroup.Add(UpgradeDone);
                cmdGroup.Add(Investigate);
                cmdGroup.Add(ReplayDataLoading);
            }

            return cmdGroup.ToArray();
        }

        /// <summary>根據編碼，解析出相對應的命令物件。</summary>
        /// <param name="code></param>
        /// <returns></returns>
        public static Command Parse(byte code)
        {
            foreach (Command cmd in GetList())
            {
                if (cmd.Code == code) {
                    return cmd;
                }
            }

            return null;
        }

        /// <summary>根據腳本型態，解析出相對應的命令物件。</summary>
        /// <param name="scriptType"></param>
        /// <returns></returns>
        public static Command Parse(ScriptType scriptType)
        {
            foreach (Command cmd in GetList())
            {
                if (cmd.ScriptType == scriptType) {
                    return cmd;
                }
            }

            return null;
        }
    }
}
