﻿using Net.Core.Communication;
using System;

namespace PH.Bootloader.RS485.DSP.Protocol
{
    /// <summary>代表封包產生、解析之類別。</summary>
    class PacketFactory
    {
        #region Fields
        private const int   PACKET_HEADER_BYTE          = 0;
        private const int   PACKET_SOURCE_BYTE          = 1;
        private const int   PACKET_DESTINATION_BYTE     = 2;
        private const int   PACKET_DATA_LENGTH_BYTE     = 3;
        private const int   PACKET_COMMAND_BYTE         = 4;
        private const int   PACKET_PAYLOAD_BYTE         = 5;
        private const byte  PACKET_HEADER               = 0xF0;
        private const byte  PACKET_INFO_SIZE            = 7;

        private const byte  TX_PACKET_SOURCE            = 0x10;
        private const uint  RX_BUFFER_SIZE              = 24;

        private readonly byte[]         rxBuffer = null;
        private static PacketFactory    singleton = null;
        #endregion

        #region Constructor
        /// <summary>初始化 PacketFactory 類別的新執行個體。</summary>
        private PacketFactory()
        {
            rxBuffer = new byte[RX_BUFFER_SIZE];
        }
        #endregion

        #region Methods
        /// <summary>取得實體類別。</summary>
        /// <param name="sourceID"></param>
        /// <returns></returns>
        public static PacketFactory GetInstance()
        {
            if (singleton == null) {
                singleton = new PacketFactory();
            }

            return singleton;
        }

        /// <summary>產生資料封包。</summary>
        /// <param name="deviceInfo"></param>
        /// <param name="cmd"></param>
        /// <param name="dataArray"></param>
        /// <param name="length"></param>
        /// <returns></returns>
        public byte[] Create(DeviceInformation deviceInfo, Command cmd, byte[] dataArray, int length)
        {
            byte len = (byte)(PACKET_INFO_SIZE + length);
            byte[] txBuffer = new byte[len];
            
            txBuffer[PACKET_HEADER_BYTE] = PACKET_HEADER;
            txBuffer[PACKET_SOURCE_BYTE] = TX_PACKET_SOURCE;
            txBuffer[PACKET_DESTINATION_BYTE] = deviceInfo.CorrespondDspID;
            txBuffer[PACKET_DATA_LENGTH_BYTE] = len;
            txBuffer[PACKET_COMMAND_BYTE] = cmd.Code;
            if (length > 0) {
                Array.Copy(dataArray, 0, txBuffer, PACKET_PAYLOAD_BYTE, length);
            }
            int crc = CRC16.Calculate(txBuffer, len - 2);
            txBuffer[len - 2] = (byte)(crc & 0xFF);
            txBuffer[len - 1] = (byte)(crc >> 8);

            return txBuffer;
        }

        /// <summary>解析接收到的封包內容。</summary>
        /// <param name="packetArray">完整的接收資料封包位元組。</param>
        /// <param name="length">接收資料封包長度。</param>
        /// <param name="content">解析後的封包物件。</param>
        /// <returns>是否解析成功。</returns>
        public bool Parse(byte[] packetArray, int length, out ReceivedPacketContent content)
        {
            bool isSuccess = false;
            content = null;

            if (length >= PACKET_INFO_SIZE)
            {
                content = new ReceivedPacketContent();
                content.Header = packetArray[PACKET_HEADER_BYTE];
                content.Source = packetArray[PACKET_SOURCE_BYTE];
                content.Destination = packetArray[PACKET_DESTINATION_BYTE];
                content.DataLength = packetArray[PACKET_DATA_LENGTH_BYTE];
                content.Command = Command.Parse(packetArray[PACKET_COMMAND_BYTE]);
                if (content.Command == null) {
                    return false;
                }

                int dataLen = length - PACKET_INFO_SIZE;
                if (dataLen > 0)
                {
                    byte[] dataArray = new byte[dataLen];
                    Array.Copy(packetArray, PACKET_PAYLOAD_BYTE, dataArray, 0, dataLen);
                    content.Payload.AddRange(dataArray);
                }
                content.ReceivedCRC = (packetArray[length - 1] << 8);
                content.ReceivedCRC += packetArray[length - 2];
                content.CalculatedCRC = CRC16.Calculate(packetArray, length - 2);
                isSuccess = true;
            }

            return isSuccess;
        }

        /// <summary>確認回應的資料內容格式是否正確。</summary>
        /// <param name="targetPSU">預期 PSU 型態。</param>
        /// <param name="targetCmd">預期指令型態。</param>
        /// <param name="content">接收資料內容。</param>
        /// <returns></returns>
        public bool ValidateResponsePacket(DeviceInformation targetPSU, Command targetCmd, ReceivedPacketContent content)
        {
            bool isSuccess = false;

            if (content.Header == PACKET_HEADER &&
                    content.Source == targetPSU.CorrespondDspID &&
                    content.Destination == TX_PACKET_SOURCE &&
                    content.DataLength == targetCmd.ResponsePacketLength &&
                    content.Command.Equals(targetCmd) &&
                    content.CalculatedCRC == content.ReceivedCRC)
            {
                isSuccess = true;
            }

            return isSuccess;
        }

        /// <summary>確認回應的資料內容格式是否正確。</summary>
        /// <param name="targetPSU">預期 PSU 型態。</param>
        /// <param name="targetCmd">預期指令型態。</param>
        /// <param name="replyCode">預期指令型態。</param>
        /// <param name="content">接收資料內容。</param>
        /// <returns></returns>
        public bool ValidateResponsePacket(DeviceInformation targetPSU, Command targetCmd, Command replyCmd, ReceivedPacketContent content)
        {
            bool isSuccess = false;

            if (content.Header == PACKET_HEADER &&
                    content.Source == targetPSU.CorrespondDspID &&
                    content.Destination == TX_PACKET_SOURCE &&
                    content.DataLength == targetCmd.ResponsePacketLength &&
                    content.Command.Equals(replyCmd) &&
                    content.CalculatedCRC == content.ReceivedCRC)
            {
                isSuccess = true;
            }

            return isSuccess;
        }
    }
    #endregion
}
