﻿using System;
using System.IO;
using System.Threading;
using Net.Core.Communication;
using Net.Core.Communication.Serial.Sync;

namespace PH.Bootloader.RS485.PIC32
{
    /// <summary>封裝 Microchip PIC32 啟動程序安裝之類別。</summary>
    public class PicBootloader : IBootloader
    {
        #region Fields
        private const int       CHECK_DEVICE_DELAY  = 800;   // 單位：毫秒
        private const int       RESET_DEVICE_DELAY  = 500;   // 單位：毫秒
        private const byte      EEP_FILE_NONE       = 0;
        private const byte      EEP_FILE_EXISTS     = 11;

        // Communication
        private ISyncCommunicationPlus  comPort;
        private bool            isComReady;
        // Command
        private string[]        hexMemory;
        private byte[]          checkCmdArray;
        private byte[]          csuResetCmdArray;
        private byte[]          psuResetCmdArray;
        private byte[]          earseCmdArray;
        // Boot procedure
        private float           progressPercentage;
        private float           progressStep;
        private PicInformation  picInfo;
        private bool            isAbort;
        // Memory
        private long            sumMemorySize;
        #endregion

        #region Properties
        /// <summary>取得、設定名稱。</summary>
        public string           Name { get; set; }
        /// <summary>取得程序執行中旗標。</summary>
        public bool             IsProcedureRunning { get; private set; }
        /// <summary>設定、取得是否由 PIC 擔任代理燒錄的腳色 (透過 CAN Bus) (用不到)。</summary>
        public bool             IsCanAgent { get; set; }
        #endregion

        #region Constructor
        /// <summary>初始化 Bootloader 類別之執行個體。</summary>
        public PicBootloader()
        {
            hexMemory = null;

            checkCmdArray = new byte[1] { 0xC1 };
            csuResetCmdArray = new byte[11] { 0xF0, 0x10, 0xFE, 0x03, 0x00, 0x11, 0x0D, 0x00, 0xAA, 0x00, 0x00 };
            psuResetCmdArray = new byte[10] { 0xF0, 0x10, 0xFE, 0x05, 0x01, 0x8A, 0xFF, 0x00, 0x00, 0x00 };
            earseCmdArray = new byte[1] { 0x45 };

            picInfo = new PicInformation();
            picInfo.CodeID = 0xD6;
            picInfo.Family = "F";
            picInfo.Description = "PIC32MX7xx w/512KB flash";
            picInfo.FlashMem = 0x80000;
            picInfo.BootloaderSize = 0x1000;
            picInfo.BlockSize = 128;
            picInfo.EepromFlag = EEP_FILE_NONE;
        }
        #endregion

        #region Methods : Public
        /// <summary>終止所有處理程序。</summary>
        public void Abort()
        {
            isAbort = true;
        }

        /// <summary>指定『已開啟』之串列通訊埠。</summary>
        /// <param name="comPort">實體串列通訊埠。</param>
        /// <returns>串列通訊埠是否已開啟。</returns>
        public bool AssignComPort(ISyncCommunicationPlus comPort)
        {
            this.comPort = comPort;
            isComReady = comPort.IsOpen;
            return isComReady;
        }
        
        /// <summary>檢查 App Code 是否存在。</summary>
        /// <param name="device"></param>
        /// <returns></returns>
        public bool CheckAppCode(DeviceInformation device)
        {
            if (!CheckSerialPort()) {
                return false;
            }

            byte id = 0xFF;
            while (DeviceInformation.ParseBy485ID(id) != null && id > 0) {
                id--;
            }

            return !ImplementCheckDevice(device.Type, id, false);
        }

        /// <summary>檢查裝置。</summary>
        /// <param name="device"></param>
        /// <returns></returns>
        public bool CheckDevice(DeviceInformation device)
        {
            if (!CheckSerialPort()) {
                return false;
            }

            Thread checkDeviceThread = new Thread(() => CheckDeviceProcedure(device.Type, device.RS485ID, 5));
            checkDeviceThread.IsBackground = true;
            checkDeviceThread.Start();

            return true;
        }

        /// <summary>解析記憶體內容。</summary>
        /// <param name="device"></param>
        /// <param name="filePath"></param>
        public void ParseHexToMemory(DeviceInformation device, string filePath)
        {
            hexMemory = null;
            IsProcedureRunning = true;

            ShowMessageCallback("\r\nParse image file...\r\n");

            if (VerifyFileName(filePath, device))
            {
                if (File.Exists(filePath))
                {
                    if (VerifyAppImage(filePath, out hexMemory))
                    {
                        ShowMessageCallback("Image file is ready!!\r\n");
                        ParseHexCallback(hexMemory);
                        CompleteRunWorkCallback(ProcedureType.HexToMemory, BootloadResult.ImageFileReady);
                    }
                    else
                    {
                        ShowMessageCallback("           Err: Image file is not an application code!!\r\n");
                        CompleteRunWorkCallback(ProcedureType.HexToMemory, BootloadResult.ImageFileNotAppCode);
                    }
                }
                else
                {
                    ShowMessageCallback("           Err: Image file dose not exist!!\r\n");
                    CompleteRunWorkCallback(ProcedureType.HexToMemory, BootloadResult.ImageFileNotExist);
                }
            }
            else
            {
                ShowMessageCallback("           Err: Image file is not match!!\r\n");
                CompleteRunWorkCallback(ProcedureType.HexToMemory, BootloadResult.ImageFileNotMatch);
            }

            IsProcedureRunning = false;
        }

        /// <summary>上載記憶體資訊。</summary>
        /// <param name="device">指定的裝置。</param>
        /// <returns>程序是否啟動。</returns>
        public bool Upload(DeviceInformation device)
        {
            if (!CheckSerialPort()) {
                return false;
            }

            if (hexMemory != null)
            {
                sumMemorySize = 0;
                Thread writeThread = new Thread(() => LoadImageProcedure(device.Type, device.RS485ID, hexMemory));
                writeThread.IsBackground = true;
                writeThread.Name = "UploadImageProcess_PIC";
                writeThread.Start();

                return true;
            }
            else
            {
                CompleteRunWorkCallback(ProcedureType.HexToMemory, BootloadResult.ImageFileNotExist);
                return false;
            }
        }
        #endregion

        #region Methods : Private : Check Device
        /// <summary>檢查串列埠是否備妥。</summary>
        /// <returns></returns>
        private bool CheckSerialPort()
        {
            if (!isComReady) {
                CompleteRunWorkCallback(ProcedureType.CheckCommunication, BootloadResult.SerialPortNotReady);
                ShowMessageCallback("           Err: Serial port is not ready yet!!\r\n");            } else
            {
                CompleteRunWorkCallback(ProcedureType.CheckCommunication, BootloadResult.SerialPortReady);
            }

            return isComReady;
        }

        /// <summary>檢查裝置程序。</summary>
        /// <param name="devType"></param>
        /// <param name="devID"></param>
        /// <param name="retryCount"></param>
        private void CheckDeviceProcedure(DeviceType devType, byte devID, int retryCount)
        {
            progressStep = 100 / retryCount;
            progressPercentage = 0;
            int retry = retryCount;

            IsProcedureRunning = true;
            while (retry > 0)
            {
                if (isAbort || ImplementCheckDevice(devType, devID, true)) {
                    break;
                }

                retry--;
                progressPercentage += progressStep;
                ChangeProcedureProgressCallback(ProcedureType.CheckDevice, (int)progressPercentage);

                Thread.CurrentThread.Join(CHECK_DEVICE_DELAY);
            }
            ChangeProcedureProgressCallback(ProcedureType.CheckDevice, 0);

            if (retry == 0)
            {
                ShowMessageCallback("           Err: No device detected!!\r\n" +
                                    "                  Verify - COM number\r\n" +
                                    "                            - baud rate\r\n" +
                                    "                            - if Device was reseted less than 1sec ago\r\n" +
                                    "                              (do you need to check \"RST Reset\"?)\r\n" +
                                    "                            - if Bootloader firmware is correctly\r\n" +
                                    "                              programmed into Device\r\n");
            }
            else if (isAbort)
            {
                ShowMessageCallback("           Abort: No device detected!!\r\n");
            }
            //ShowMessageCallback("--------------------------------------------------------------\r\n");

            IsProcedureRunning = false;
            isAbort = false;
            CompleteRunWorkCallback(ProcedureType.CheckDevice, retry == 0 ? BootloadResult.DetectNothing : BootloadResult.DetectDevice);
        }

        /// <summary>實作檢查裝置。</summary>
        /// <param name="devType"></param>
        /// <param name="devID"></param>
        /// <param name="isShowMsg"></param>
        /// <returns></returns>
        private bool ImplementCheckDevice(DeviceType devType, byte devID, bool isShowMsg)
        {
            ResetDevice(devType, devID, isShowMsg);
            return SendBootKey(isShowMsg);
        }

        /// <summary>重置裝置。</summary>
        /// <param name="devType"></param>
        /// <param name="devID"></param>
        /// <param name="isShowMsg"></param>
        private void ResetDevice(DeviceType devType, byte devID, bool isShowMsg)
        {
            byte[] outComePtr;
            int crc16;

            if (devType.Equals(DeviceType.PIC_CSU) ||
                devType.Equals(DeviceType.PIC_VCU) ||
                devType.Equals(DeviceType.PIC_SBox))
            {
                csuResetCmdArray[2] = devID;
                crc16 = CRC16.Calculate(csuResetCmdArray, 9);
                csuResetCmdArray[9] = (byte)((crc16 & 0x00FF));        
                csuResetCmdArray[10] = (byte)((crc16 & 0xFF00) >> 8);

                comPort.SendExpectedLengthReceive(csuResetCmdArray, 11, 10, out outComePtr, true);
            }
            else if (devType.Equals(DeviceType.PIC_PSU) ||
                     devType.Equals(DeviceType.PIC_PFC))
            {
                psuResetCmdArray[2] = devID;
                crc16 = CRC16.Calculate(psuResetCmdArray, 8);
                psuResetCmdArray[8] = (byte)((crc16 & 0x00FF));
                psuResetCmdArray[9] = (byte)((crc16 & 0xFF00) >> 8);

                comPort.SendExpectedLengthReceive(psuResetCmdArray, 10, 10, out outComePtr, true);
            }

            if (isShowMsg) {
                ShowMessageCallback("\r\nWait MCU Reset...\r\n");
            }
            
            Thread.CurrentThread.Join(RESET_DEVICE_DELAY);
        }

        /// <summary>傳送 Boot 代碼。</summary>
        /// <param name="isShowMsg"></param>
        /// <returns></returns>
        private bool SendBootKey(bool isShowMsg)
        {
            if (isShowMsg) {
                ShowMessageCallback("Check Device...\r\n");
                ShowMessageCallback("           Send: 0xC1\r\n");
            }
            
            byte[] outComePtr;
            if (comPort.SendExpectedLengthReceive(checkCmdArray, 1, 2, out outComePtr))
            {
                byte familyCode = outComePtr[1];
                if (isShowMsg) {
                    ShowMessageCallback(string.Format("           Receive: 0x{0}, 0x{1}\r\n", outComePtr[0].ToString("X2"), outComePtr[1].ToString("X2")));
                }

                if ((familyCode < 0x30) || ((familyCode > 0x39) && 
                    (familyCode < 0x41)) || (familyCode > 0x5A))
                {
                    if (isShowMsg) {
                        ShowMessageCallback("           Err: Wrong answer from device!!\r\n");
                    }   
                }
                else
                {
                    picInfo.Family = (familyCode == 0x6D || familyCode == 0x4B) ? 
                                    "F" : Char.ConvertFromUtf32(familyCode);

                    if (isShowMsg) {
                        ShowMessageCallback(picInfo.CodeID != 0 ?
                            string.Format("           MCU:  {0}\r\n", picInfo.Description) :
                            "           Err: Unknown device ID code and/or family\r\n");
                    }    
                }

                return true;
            }

            return false;
        }
        #endregion

        #region Methods : Private : Upload
        /// <summary>確認檔案名稱。</summary>
        /// <param name="filePath"></param>
        /// <param name="deviceInfo"></param>
        /// <returns></returns>
        private bool VerifyFileName(string filePath, DeviceInformation deviceInfo)
        {
            if (!string.IsNullOrEmpty(filePath))
            {
                string strFileName = Path.GetFileName(filePath).Substring(0, 3);
                string strTargetName = deviceInfo.Type.HexKey.Substring(0, 3);
                return (string.Compare(strFileName, strTargetName, StringComparison.OrdinalIgnoreCase) == 0);
            }

            return false;
        }

        /// <summary>確認映像檔為 App Code。</summary>
        /// <param name="filePath"></param>
        /// <param name="lines"></param>
        /// <returns></returns>
        private bool VerifyAppImage(string filePath, out string[] lines)
        {
            bool isVaild = false;

            lines = File.ReadAllLines(filePath);
            if (lines != null && lines.Length >= 30)
            {
                for (int i = 0; i < 30; i++)
                {
                    if (lines[i].Length == 43)
                    {
                        isVaild = string.Equals(lines[i].Substring(0, 7), ":107000");
                        if (isVaild) {
                            break;
                        }
                    }
                }
            }

            return isVaild;
        }

        /// <summary>載入映像檔程序。</summary>
        /// <param name="devType"></param>
        /// <param name="devID"></param>
        /// <param name="lines"></param>
        private void LoadImageProcedure(DeviceType devType, byte devID, string[] lines)
        {
            byte[] flashMemArray;
            IsProcedureRunning = true;

            if (ImplementCheckDevice(devType, devID, true))
            {
                CompleteRunWorkCallback(ProcedureType.UploadCheckDevice, BootloadResult.DetectDevice);

                if (ParseImageToMemoryByte(lines, out flashMemArray))
                {
                    CompleteRunWorkCallback(ProcedureType.HexToMemory, BootloadResult.MemoryAllocationReady);

                    ConfigureMemoryParameter();

                    if (EraseFlash())
                    {
                        CompleteRunWorkCallback(ProcedureType.HexToMemory, BootloadResult.FlashErased);
                        ChangeUploadInformationCallback(0, picInfo.MemPosMax);

                        if (LoadImage(devType, devID, flashMemArray))
                        {
                            CompleteRunWorkCallback(ProcedureType.Upload, BootloadResult.Uploaded);
                            ShowMessageCallback("Transferred program memory successfully ★★★★★\r\n");
                        }
                        else
                        {
                            CompleteRunWorkCallback(ProcedureType.Upload, BootloadResult.UploadFailed);
                            ShowMessageCallback(isAbort ? "           Abort: No device detected!!\r\n" :
                                                          "           Err: Problem in writing flash memory!!\r\n");
                        }
                    }
                    else
                    {
                        CompleteRunWorkCallback(ProcedureType.HexToMemory, BootloadResult.EraseFlashFailed);
                        ShowMessageCallback(isAbort ? "           Abort: No device detected!!\r\n" :
                                                      "           Err: Erasing flash memory failed!!\r\n");
                    }
                }
                else
                {
                    CompleteRunWorkCallback(ProcedureType.HexToMemory, BootloadResult.MemoryAllocationFailed);
                    ShowMessageCallback("           Err: Parse image failed!!\r\n");
                }
            }
            else
            {
                CompleteRunWorkCallback(ProcedureType.UploadCheckDevice, BootloadResult.DetectNothing);
                ShowMessageCallback("           Err: No device detected!!\r\n");
            }

            IsProcedureRunning = false;
            isAbort = false;
        }

        /// <summary>解析映像檔為位元組資料。</summary>
        /// <param name="lines"></param>
        /// <param name="memArray"></param>
        /// <returns></returns>
        private bool ParseImageToMemoryByte(string[] lines, out byte[] memArray)
        {
            memArray = new byte[picInfo.FlashMem];
            for (int i = 0; i< picInfo.FlashMem; i++) {
                memArray[i] = 0xFF;
            }

            int address = 0;
            int extendedAddress = 0;        // extented address for INHX32 hex files (PIC18 & PIC32)
            int byteCountPerLine = 0;
            int recordType = 0;
            bool isCodeSection = true;      // if we are in code section so transfert byte from hex to array
            bool isBootSection = false;     // only for PIC32, if we are in boot flash memory
            bool hasPIC32JumpMoved = false; // only for PIC32, boot section has priority to move the jump from reset vector to specified place
            bool hasData = false;

            try
            {
                picInfo.TotalByteInFile = 0;

                foreach (string line in lines)
                {
                    if (line.Substring(0, 1).Trim() == ":")
                    {
                        address = Convert.ToInt32(line.Substring(3, 4).Trim(), 16);
                        recordType = Convert.ToInt32(line.Substring(7, 2).Trim(), 16);

                        if ((recordType == 0) && (isCodeSection || isBootSection))
                        {
                            byteCountPerLine = Convert.ToInt32(line.Substring(1, 2).Trim(), 16);

                            for (int i = 9; i < (9 + 2 * byteCountPerLine); i += 2)
                            {
                                if (isCodeSection)
                                {
                                    if (((extendedAddress + address) >= 0x6000) &&
                                        ((extendedAddress + address) < 0x6010) &&
                                        (picInfo.Family == "F") &&
                                        (hasPIC32JumpMoved == false))
                                    {
                                        memArray[(extendedAddress + address - 0x6000) + picInfo.FlashMem - (picInfo.BootloaderSize + 0x10)] = Convert.ToByte(line.Substring(i, 2).Trim(), 16);
                                    }

                                    if (picInfo.EepromFlag != EEP_FILE_EXISTS) {
                                        memArray[extendedAddress + address] = Convert.ToByte(line.Substring(i, 2).Trim(), 16);
                                    }

                                    picInfo.TotalByteInFile++;
                                }

                                address++;
                            }

                            hasData = true;
                        }
                        else if (recordType == 4)
                        {
                            extendedAddress = Convert.ToInt32(line.Substring(9, 4).Trim(), 16);

                            // for PIC32:
                            // address extension = 0x1D00 to 0x1D07 (PIC32 program flash mem), reserved from 0x1D08 to 0x1F80 (if one day memory > to 512kB) 
                            if ((extendedAddress >= 0x1D00) && (extendedAddress < 0x1E00))
                            {
                                isCodeSection = true;
                                isBootSection = false;
                                extendedAddress = ((extendedAddress & 0x00FF) << 16);
                            }
                            // for PIC32:
                            if (extendedAddress == 0x1FC0)
                            {
                                isCodeSection = false;
                                isBootSection = true;
                                extendedAddress = 0; // not really used
                            }
                        }
                    }
                }
            }
            catch (Exception)
            {
                return false;
            }

            return hasData;
        }

        /// <summary>配置記憶體參數。</summary>
        private void ConfigureMemoryParameter()
        {
            if (picInfo.Family == "F")
            {
                picInfo.DividerHexToMemory = 1;
                picInfo.MemPosMax = picInfo.FlashMem - picInfo.BootloaderSize;
                picInfo.MemPosMin = 0;
            } 
        }

        /// <summary>抹除記憶體。</summary>
        /// <returns></returns>
        private bool EraseFlash()
        {
            byte totalPage = 0;
            byte erasedPage = 0;

            try
            {
                if (picInfo.Family.Equals("F"))
                {
                    ShowMessageCallback("\r\nErasing memory...\r\n");
                    
                    comPort.Send(earseCmdArray, 1);

                    totalPage = comPort.ReadByte();
                    progressStep = 100.0F / totalPage;
                    progressPercentage = 0;

                    while (comPort.ReadByte() != 0x4B)
                    {
                        progressPercentage += progressStep;
                        ChangeProcedureProgressCallback(ProcedureType.HexToMemory, (int)progressPercentage);

                        erasedPage++;
                        if (erasedPage > totalPage) {
                            ChangeProcedureProgressCallback(ProcedureType.HexToMemory, 0);
                            return false;
                        }
                    }

                    ChangeProcedureProgressCallback(ProcedureType.HexToMemory, 0);
                    ShowMessageCallback("Erased memory successfully!!\r\n");
                    return true;
                } 
            }
            catch
            {
            }

            return false;
        }

        /// <summary>傳送檔案。</summary>
        /// <param name="devType"></param>
        /// <param name="deviceID"></param>
        /// <param name="memArray"></param>
        /// <returns></returns>
        private bool LoadImage(DeviceType devType, byte deviceID, byte[] memArray)
        {
            if (picInfo.EepromFlag != EEP_FILE_EXISTS)
            {
                ShowMessageCallback("\r\nTransferring program memory...\r\n");

                progressStep = (picInfo.BlockSize * 100.0F) / picInfo.TotalByteInFile;
                progressPercentage = 0;
                sumMemorySize = 0;

                for (int i = picInfo.MemPosMin; i < picInfo.MemPosMax; i += picInfo.BlockSize)
                {
                    if (!PrepareAndWriteBlock(i, memArray)) {
                        ChangeProcedureProgressCallback(ProcedureType.Upload, 0);
                        ChangeUploadInformationCallback(0, 0);
                        return false;
                    }
                }
            }

            ChangeProcedureProgressCallback(ProcedureType.Upload, 0);
            ChangeUploadInformationCallback(0, 0);
            return true;
        }

        /// <summary>準備資料區塊及寫入。</summary>
        /// <param name="memPos"></param>
        /// <param name="memArray"></param>
        /// <returns></returns>
        private bool PrepareAndWriteBlock(int memPos, byte[] memArray)
        {
            try
            {
                int sizeBlock = picInfo.BlockSize;
                int nonCount = 0;
                int i = 0;

                byte[] memBlock = new byte[sizeBlock];
                for (i = 0; i < sizeBlock; i++) {
                    memBlock[i] = 0xFF;
                }

                for (i = 0; i < sizeBlock; i++) {
                    if (memArray[memPos + i] == 0xFF) {
                        nonCount++;
                    } else {
                        memBlock[i] = memArray[memPos + i];
                    }
                }

                // 全部為 0xFF 則不進行寫入
                if (nonCount < sizeBlock)
                {
                    int retry = 5;
                    while (!isAbort && retry > 0 && !WriteBlock(memPos, memBlock, 0)) {
                        retry--;
                    }

                    progressPercentage += progressStep;
                    ChangeProcedureProgressCallback(ProcedureType.Upload, (int)progressPercentage);

                    // 2018/12/5
                    // 進度條與資料進度不一致!!這算作弊.....@@
                    sumMemorySize += sizeBlock;
                    if (sumMemorySize > picInfo.TotalByteInFile) {
                        sumMemorySize = picInfo.TotalByteInFile;
                    }
                    ChangeUploadInformationCallback(sumMemorySize, picInfo.TotalByteInFile);

                    return (!isAbort) && (retry > 0);
                }

                return true;
            }
            catch 
            {
                return false;
            }
        }

        /// <summary>寫入區塊資料。</summary>
        /// <param name="memPos"></param>
        /// <param name="memBlock"></param>
        /// <param name="option">0 = program flash memory; 1 = boot</param>
        /// <returns></returns>
        private bool WriteBlock(int memPos, byte[] memBlock, byte option)
        {
            byte sizeBlock = (byte)memBlock.Length;
            byte[] checkSum = new byte[2];
            byte[] infoArray = new byte[5];
            byte uByte = (byte)(((memPos / picInfo.DividerHexToMemory) >> 16) & 0xFF);
            byte hByte = (byte)(((memPos / picInfo.DividerHexToMemory) >> 8) & 0xFF);
            byte lByte = (byte)((memPos / picInfo.DividerHexToMemory) & 0xFF);

            int crc16 = CRC16.Calculate(memBlock, sizeBlock);
            checkSum[0] = (byte)((crc16 & 0x00FF));
            checkSum[1] = (byte)((crc16 & 0xFF00) >> 8);
            checkSum[0] += (byte)(uByte + hByte + lByte + sizeBlock + option);

            int lenPackage = picInfo.Family == "F" ? (infoArray.Length + sizeBlock + checkSum.Length) : (sizeBlock + checkSum.Length);
            byte[] packageArray = new byte[lenPackage];

            // PIC32 family
            if (picInfo.Family == "F")
            {
                infoArray[0] = sizeBlock;
                infoArray[1] = option;
                infoArray[2] = uByte;
                infoArray[3] = hByte;
                infoArray[4] = lByte;
                Array.Copy(infoArray, 0, packageArray, 0, infoArray.Length);
                Array.Copy(memBlock, 0, packageArray, infoArray.Length, memBlock.Length);
                Array.Copy(checkSum, 0, packageArray, infoArray.Length + memBlock.Length, checkSum.Length);
            }
            else
            {
                Array.Copy(memBlock, 0, packageArray, 0, memBlock.Length);
                Array.Copy(checkSum, 0, packageArray, memBlock.Length, checkSum.Length);
            }

            try
            {
                comPort.Send(packageArray, packageArray.Length);
                byte receValue = comPort.ReadByte();

                /*if (receValue >= 0x4B)
                {
                    checkSum[0] = 0;
                    return true;
                }
                else if (receValue == 0x4E) // 0x4E = "N"
                {
                    return false;
                }*/
                if (receValue == 0x4B)
                {
                    checkSum[0] = 0;
                    return true;
                }
                else
                {
                    return false;
                }
            }
            catch
            {
                ShowMessageCallback("           Err: No answer while Transferring block data!!\r\n");
                return false;
            }
        }
        #endregion

        #region Methods : Callback
        private void ParseHexCallback(string[] records)
        {
            HexParsed?.Invoke(this, records);
        }

        private void ShowMessageCallback(string message)
        {
            MessageShown?.Invoke(this, message);
        }

        private void CompleteRunWorkCallback(ProcedureType procType, BootloadResult result)
        {
            ProcedureCompleted?.Invoke(this, procType, result);
        }

        private void ChangeProcedureProgressCallback(ProcedureType procType, int progressPercentage)
        {
            if (progressPercentage < 0){
                progressPercentage = 0;
            } else if (progressPercentage > 100) {
                progressPercentage = 100;
            }

            ProcedureProgressChanged?.Invoke(this, procType, progressPercentage);
        }

        private void ChangeUploadInformationCallback(long current, long total)
        {
            UploadInformationChanged?.Invoke(this, current, total);
        }
        #endregion

        #region Events
        /// <summary>發生在解析出 Hex 內容後。</summary>
        public event HexParsedEventHandler                  HexParsed;
        /// <summary>發生在更新程序信息顯示時。</summary>
        public event MessageShownEventHandler               MessageShown;
        /// <summary>發生在程序執行完成時。</summary>
        public event ProcedureCompletedEventHandler         ProcedureCompleted;
        /// <summary>發生在程序進度改變時。</summary>
        public event ProcedureProgressChangedEventHandler   ProcedureProgressChanged;
        /// <summary>發生在資料更新改變時。</summary>
        public event UploadInformationChangedEventHandler   UploadInformationChanged;
        #endregion
    }
}
