﻿using System;
using System.IO;
using System.Linq;
using PH.Bootloader.RS485.PIC32_2.Protocol;

namespace PH.Bootloader.RS485.PIC32_2
{
    /// <summary>封裝 Microchip PIC32 啟動程序更新之類別(2.0版/瑱賢)。</summary>
    public partial class PIC32Bootloader2
    {
        /// <summary>清除記憶體內容。</summary>
        public void Clear()
        {
            upgradeMemHandler.Clear();
        }

        /// <summary>載入 HEX 檔。</summary>
        /// <param name="deviceInfo">裝置資訊。</param>
        /// <param name="filePath">映像檔案路徑。</param>
        /// <param name="imgType">更新程式碼型態。</param>
        /// <returns>解析失敗/沒有資料。</returns>
        public bool LoadHexFile(DeviceInformation deviceInfo, string filePath, UpgradeImageType imgType)
        {
            if (File.Exists(filePath))
            {
                string[] strLines;
                if (ReadFileLines(filePath, out strLines)) {
                    if (ParseLinesToMemory(strLines, imgType)) {
                        return VerifyMemoryInfo(deviceInfo.Type, imgType);
                    } else {
                        ChangeUpgradeStatusCallback(UpgradeStatusType.ImageFileEmpty);
                    }
                }
            }
            else
            {
                ChangeUpgradeStatusCallback(UpgradeStatusType.ImageFileNotExist);
            }

            return false;
        }

        private bool ReadFileLines(string filePath, out string[] lines)
        {
            lines = File.ReadAllLines(filePath);
            return lines != null && lines.Length > 0;
        }

        private bool ParseLinesToMemory(string[] strLines, UpgradeImageType upgradeType)
        {
            upgradeMemHandler.Clear();

            // 預留 2KB 的記憶體空間，透過 HEX 解析，計算出實際的記憶體大小
            // HEX 解析規則根據：Intel HEX Standard
            int i = 0, address = 0, recLen = 0, recType = 0, extAddress = 0;
            bool hasData = false;

            try
            {
                // Each record begins with a 『Record Mark』
                strLines.Select(s => s.Substring(0, 1).Trim() == ":");

                upgradeMemHandler.Clear();

                foreach (string strLine in strLines)
                {
                    recLen = Convert.ToInt32(strLine.Substring(1, 2).Trim(), 16);
                    address = Convert.ToInt32(strLine.Substring(3, 4).Trim(), 16);
                    recType = Convert.ToInt32(strLine.Substring(7, 2).Trim(), 16);

                    if (recType == 0)   // Data Record
                    {
                        address = extAddress + address;

                        for (i = 9; i < (9 + 2 * recLen); i += 2) {
                            if (upgradeMemHandler.FillMemory(address, Convert.ToByte(strLine.Substring(i, 2).Trim(), 16), upgradeType)) {
                                address++;
                            } else {
                                ChangeUpgradeStatusCallback(UpgradeStatusType.ImageFileNotMatch);
                                return false;
                            }
                        }

                        hasData = true;
                    }
                    else if (recType == 4)  // Extended Linear Address Record
                    {
                        // PIC32 AppCode 範圍: 0x1D00~0x1D07
                        // 預留接收其他型號的 AppCode (如果記憶體空間 > 512KB)
                        extAddress = Convert.ToInt32(strLine.Substring(9, 4).Trim(), 16);
                        extAddress = extAddress << 16;
                    }
                }
            }
            catch (Exception)
            {
                ChangeUpgradeStatusCallback(UpgradeStatusType.MemoryAllocationFailed);
                return false;
            }

            return hasData;
        }

        private bool VerifyMemoryInfo(DeviceType devType, UpgradeImageType imgType)
        {
            bool isValidSign = false;
            bool isValidCRC = false;

            if (imgType == UpgradeImageType.AppCode ||
                imgType == UpgradeImageType.AppCodeByBootCode)
            {
                isValidSign = upgradeMemHandler.VerifyAppCodeSign(devType);
                if (isValidSign) {
                    isValidCRC = upgradeMemHandler.VerifyAppCodeCheckSum();
                } 
            }
            else if (imgType == UpgradeImageType.BootCode)
            {
                isValidSign = upgradeMemHandler.VerifyBootCodeSign(devType);
                if (isValidSign) {
                    isValidCRC = upgradeMemHandler.VerifyBootCodeCheckSum();
                }    
            }

            if (!isValidSign) {
                ChangeUpgradeStatusCallback(UpgradeStatusType.ImageFileNotMatch);
            } else if (!isValidCRC) {
                ChangeUpgradeStatusCallback(UpgradeStatusType.MemoryInvalidCRC);
            }

            if (isValidSign && isValidCRC) {
                ChangeUpgradeStatusCallback(UpgradeStatusType.ImageFileReady);
            }

            return isValidSign && isValidCRC;
        }
    }
}
