﻿using System;
using System.Threading;
using Net.Core.Communication;
using PH.Bootloader.RS485.PIC32_2.Protocol;

namespace PH.Bootloader.RS485.PIC32_2
{
    /// <summary>封裝 Microchip PIC32 啟動程序更新之類別(2.0版/瑱賢)。</summary>
    public partial class PIC32Bootloader2
    {
        /// <summary>執行升級。</summary>
        /// <param name="imgType">更新程式碼型態。</param>
        /// <returns>程序是否啟動。</returns>
        public bool Upgrade(UpgradeImageType imgType)
        {
            if (upgradeMemHandler.IsMemReady)
            {
                Thread thread = new Thread(() => UpgradeMemoryProcess(imgType));
                thread.IsBackground = true;
                thread.Name = "UpgradeMemoryProcess";
                thread.Start();

                return true;
            }
            else
            {
                ChangeUpgradeStatusCallback(UpgradeStatusType.MemoryNotAllocated);
            }

            return false;
        }

        private void UpgradeMemoryProcess(UpgradeImageType imgType)
        {
            if (TriggerDeviceProcedure())
            {
                if (EraseMemoryProcess())
                {
                    if (UpgradeMemory(imgType))
                    {
                        if (ExecuteVerify())
                        {
                            // 程序上必須!!
                            Thread.Sleep(200);

                            if (EraseMemoryProcess()) {
                                if (CopyFlashProcess()) {
                                    ChangeUpgradeStatusCallback(UpgradeStatusType.UpgradeFinished);
                                    return;
                                }
                            }   
                        }
                    }
                }

                ChangeUpgradeStatusCallback(UpgradeStatusType.UpgradeFailed);
            }
        }

        private bool TriggerDeviceProcedure()
        {
            int retry = 0;

            IsProcedureRunning = true;
            while (retry < TRIGGER_RETRY_MAX) {
                if (isAbort || ExecuteTriggerDevice()) {
                    break;
                }
                retry++;
                Thread.Sleep(COMMAND_RETRY_ITVL);
            }
            IsProcedureRunning = false;

            ChangeUpgradeStatusCallback(retry < TRIGGER_RETRY_MAX ? UpgradeStatusType.TriggerFinished : UpgradeStatusType.TriggerFailed);
            return retry < TRIGGER_RETRY_MAX;
        }

        private bool UpgradeMemory(UpgradeImageType upgradeType)
        {
            bool isSuccess = true;
            byte[] memBlock = new byte[upgradeMemHandler.MemBlockSize];
            int retry;

            progressStep = (float)upgradeMemHandler.MemBlockSize / (upgradeMemHandler.MemIndexMax - upgradeMemHandler.MemIndexMin);
            progressPercentage = 0;

            IsProcedureRunning = true;
            ChangeUpgradeProgressCallback(UpgradeProgressType.Upgrade, 0);
            ChangeUpgradeStatusCallback(UpgradeStatusType.Upgrading);

            int i = 0;
            int address = 0;
            byte zone = 0;
            UpgradeBlockResult result = UpgradeBlockResult.Undefine;
            for (i = upgradeMemHandler.MemIndexMin; i < upgradeMemHandler.MemIndexMax; i += upgradeMemHandler.MemBlockSize)
            {
                if (upgradeMemHandler.GetMemory(i, ref memBlock))
                {
                    progressPercentage += progressStep;
                    if (progressPercentage >= 1.0F) {
                        progressPercentage = 1.0F;
                    }
                    ChangeUpgradeProgressCallback(UpgradeProgressType.Upgrade, (int)(progressPercentage*100.0F));

                    // 整個區塊都是 0xFF　就不用傳
                    if (CheckValidBlock(memBlock))
                    {
                        retry = 0;
                        address = upgradeMemHandler.ConvertMemoryIndexToAddress(upgradeType, i);
                        zone = upgradeMemHandler.GetZoneIndex(address);
                        while (!isAbort && retry < UPGRADE_BLOCK_RETRY_MAX) {
                            result = UpgradeBlock(address, memBlock, zone);
                            if (result == UpgradeBlockResult.Success ||
                                result == UpgradeBlockResult.InvalidAddress)
                            {
                                break;
                            }
                            retry++;
                            Thread.Sleep(COMMAND_RETRY_ITVL);
                        }
                        
                        if (result != UpgradeBlockResult.Success)
                        {
                            if (result == UpgradeBlockResult.NoResponse)
                            {
                                ChangeUpgradeStatusCallback(UpgradeStatusType.UpgradeNoResponse);
                            }
                            else if (result == UpgradeBlockResult.InvalidAddress)
                            {
                                ChangeUpgradeStatusCallback(UpgradeStatusType.UpgradeInvalidAddress);
                            }
                            else if (result == UpgradeBlockResult.Undefine)
                            {
                                ChangeUpgradeStatusCallback(UpgradeStatusType.UpgradeResponseUndefine);
                            }

                            isSuccess = false;
                            break;
                        }
                    }
                }
                else
                {
                    isSuccess = false;
                    break;
                }
            }

            IsProcedureRunning = false;
            ChangeUpgradeProgressCallback(UpgradeProgressType.Upgrade, 0);

            return isSuccess;
        }

        private UpgradeBlockResult UpgradeBlock(int memAddress, byte[] memBlock, byte zone)
        {
            byte sizeBlock = (byte)memBlock.Length;
            byte uByte = (byte)((memAddress >> 16) & 0xFF);
            byte hByte = (byte)((memAddress >> 8) & 0xFF);
            byte lByte = (byte)(memAddress & 0xFF);

            int lenPackage = 5 + sizeBlock + 2;
            byte[] packageArray = new byte[lenPackage];
            packageArray[0] = sizeBlock;
            packageArray[1] = zone;
            packageArray[2] = uByte;
            packageArray[3] = hByte;
            packageArray[4] = lByte;
            Array.Copy(memBlock, 0, packageArray, 5, sizeBlock);
            int crc16 = CRC16.Calculate(memBlock, sizeBlock);
            packageArray[lenPackage - 2] = (byte)((crc16 & 0x00FF));
            packageArray[lenPackage - 1] = (byte)((crc16 & 0xFF00) >> 8);
            packageArray[lenPackage - 2] += (byte)(sizeBlock + zone + uByte + hByte + lByte);

            protocolHandler.Send(packageArray, lenPackage);

            byte ack;
            if (protocolHandler.ReadByteAtCertainTime(out ack, commSyncTime)) {
                return ParseUpgradeBlockAck(ack); 
            }

            return UpgradeBlockResult.NoResponse;
        }

        private UpgradeBlockResult ParseUpgradeBlockAck(byte ack)
        {
            if (ack == (byte)UpgradeBlockResult.Fail)
            {
                return UpgradeBlockResult.Fail;
            }
            else if (ack == (byte)UpgradeBlockResult.InvalidAddress)
            {
                return UpgradeBlockResult.InvalidAddress;
            }
            else if (ack == (byte)UpgradeBlockResult.Success)
            {
                return UpgradeBlockResult.Success;
            }
            else
            {
                return UpgradeBlockResult.Undefine;
            }
        }
    }

    enum UpgradeBlockResult : byte
    {
        Undefine        = 0x00,
        NoResponse      = 0x01,
        Fail            = 0x45,
        InvalidAddress  = 0x49,
        Success         = 0x4B,
    }
}
