﻿using PH.Bootloader.RS485.Communication;
using PH.Bootloader.RS485.PIC32_2.Protocol;

namespace PH.Bootloader.RS485.PIC32_2
{
    /// <summary>類別中，處理更新狀態改變事件的方法。</summary>
    /// <param name="sender"></param>
    /// <param name="errorType"></param>
    public delegate void UpgradeStatusChangedEventHandler(object sender, UpgradeStatusType statusType);
    /// <summary>類別中，處理更新進度改變事件的方法。</summary>
    /// <param name="sender"></param>
    /// <param name="progressType"></param>
    /// <param name="percentage"></param>
    public delegate void UpgradeProgressChangedEventHandler(object sender, UpgradeProgressType progressType, int percentage);

    /// <summary>封裝 Microchip PIC32 啟動程序更新之類別(2.0版/瑱賢)。</summary>
    public partial class PIC32Bootloader2
    {
        #region Constants
        private const int               UPGRADE_BLOCK_RETRY_MAX = 5;
        private const int               TRIGGER_RETRY_MAX       = 5;
        private const int               COMMAND_RETRY_ITVL      = 50;
        private const int               COMMUNICATION_SYNC_TIME = 3000;
        #endregion

        #region Fields
        // Communication
        private ProtocolHandler         protocolHandler;
        private bool                    isComReady;
        private int                     commSyncTime;
        // Boot procedure
        private bool                    isAbort;
        private UpgradeMemoryHandler    upgradeMemHandler;
        // Progress
        private float                   progressStep;
        private float                   progressPercentage;
        #endregion

        #region Properties
        /// <summary>取得程序執行中旗標。</summary>
        public bool                     IsProcedureRunning { get; private set; }
        #endregion

        #region Constructor
        /// <summary>初始化 PIC32Bootloader2 類別的新執行個體。</summary>
        public PIC32Bootloader2()
        {
            protocolHandler = new ProtocolHandler();
            isComReady = false;
            commSyncTime = COMMUNICATION_SYNC_TIME;

            isAbort = false;
            upgradeMemHandler = new UpgradeMemoryHandler();
        }
        #endregion

        #region Methods : Public
        /// <summary>終止所有處理程序。</summary>
        public void Abort()
        {
            isAbort = true;
        }

        /// <summary>指定『已開啟』之串列通訊埠。</summary>
        /// <param name="comPort">實體串列通訊埠。</param>
        /// <returns>串列通訊埠是否已開啟。</returns>
        public bool AssignComPort(SerialPortHandler comPort)
        {
            isComReady = comPort.IsOpen;
            return protocolHandler.AssignComPort(comPort); ;
        }

        /// <summary>確認通訊埠是否開啟。</summary>
        /// <returns>是否開啟通訊。</returns>
        public bool CheckCommlPort()
        {
            if (!isComReady) {
                ChangeUpgradeStatusCallback(UpgradeStatusType.SerialPortNotReady);
                return false;
            }

            return isComReady;
        }

        /// <summary>取得啟動程序版本。</summary>
        /// <param name="deviceInfo">裝置資訊。</param>
        /// <returns>版本號字串。</returns>
        public string GetVersion(DeviceInformation deviceInfo)
        {
            string version;
            ExecuteGetVersion(deviceInfo.RS485ID, out version);
            return version ;
        }

        /// <summary>設定通訊同步時間(傳送等待時間)。</summary>
        /// <param name="msec"></param>
        public void SetCommSyncTime(int msec)
        {
            commSyncTime = msec;
        }

        /// <summary>更新致能(AppCode)。</summary>
        /// <param name="deviceInfo">裝置資訊。</param>
        /// <param name="imgType">更新程式碼型態。</param>
        /// <returns>指令是否成功。</returns>
        public bool UpgradeEnable(DeviceInformation deviceInfo, UpgradeImageType imgType)
        {
            return ExecuteUpgradeEnable(deviceInfo.RS485ID, imgType);
        }

        /// <summary>重置(AppCode)。</summary>
        /// <param name="deviceInfo">裝置資訊。</param>
        /// <param name="imgType">更新程式碼型態。</param>
        /// <returns>指令是否成功。</returns>
        public bool Reset(DeviceInformation deviceInfo, UpgradeImageType imgType)
        {
            return ExecuteReset(deviceInfo.RS485ID, imgType);
        }
        #endregion

        #region Methods : Private
        private bool CheckValidBlock(byte[] memBlock)
        {
            int nonCount = 0;

            for (int i = 0; i < memBlock.Length; i++) {
                if (memBlock[i] == 0xFF) {
                    nonCount++;
                }
            }

            return nonCount != memBlock.Length;
        }

        private void ChangeUpgradeStatusCallback(UpgradeStatusType statusType)
        {
            UpgradeStatusChanged?.Invoke(this, statusType);
        }

        private void ChangeUpgradeProgressCallback(UpgradeProgressType progressType, int percentage)
        {
            UpgradeProgressChanged?.Invoke(this, progressType, percentage);
        }
        #endregion

        #region Events
        /// <summary>發生在錯誤信息回報時。</summary>
        public event UpgradeStatusChangedEventHandler   UpgradeStatusChanged;
        /// <summary>發生在進度改變時。</summary>
        public event UpgradeProgressChangedEventHandler UpgradeProgressChanged;
        #endregion
    }
}
