﻿using Net.Core.Communication;
using PH.Bootloader.RS485.PIC32_2.Protocol.Command.AppCode;

namespace PH.Bootloader.RS485.PIC32_2.Protocol
{
    /// <summary>應用程式碼封包之處理類別。</summary>
    class AppCodePacketFactory
    {
        private const byte  TX_PACKET_HEAD                  = 0xF0;
        private const byte  TX_PACKET_MASTER                = 0x10;
        private const int   TX_PACKET_SIZE                  = 10;

        private const int   TX_PACKET_HEAD_BYTE             = 0;
        private const int   TX_PACKET_SOURCE_BYTE           = 1;
        private const int   TX_PACKET_DESTINATION_BYTE      = 2;
        private const int   TX_PACKET_FUNCTION_BYTE         = 3;
        private const int   TX_PACKET_ADDRESS_HIGH_BYTE     = 4;
        private const int   TX_PACKET_ADDRESS_LOW_BYTE      = 5;
        private const int   TX_PACKET_QUANTITY_HIGH_BYTE    = 6;
        private const int   TX_PACKET_QUANTITY_LOW_BYTE     = 7;
        private const int   TX_PACKET_CRC_VALUE_LOW_BYTE    = 8;
        private const int   TX_PACKET_CRC_VALUE_HIGH_BYTE   = 9;

        private const int   RX_PACKET_HEAD_BYTE             = 0;
        private const int   RX_PACKET_SOURCE_BYTE           = 1;
        private const int   RX_PACKET_DESTINATION_BYTE      = 2;
        private const int   RX_PACKET_FUNCTION_BYTE         = 3;
        private const int   RX_PACKET_DATA_LENGTH_BYTE      = 4;
        private const int   RX_PACKET_DATA_START_BYTE       = 5;

        /// <summary>產生資料封包陣列。</summary>
        /// <param name="deviceID">裝置位址。</param>
        /// <param name="cmd">應用程式碼指令。</param>
        /// <returns>傳送的資料陣列。</returns>
        public static byte[] Create(byte deviceID, IAppCommand cmd)
        {
            byte[] packetArray = new byte[TX_PACKET_SIZE];
            packetArray[0] = TX_PACKET_HEAD;
            packetArray[1] = TX_PACKET_MASTER;
            packetArray[2] = deviceID;
            packetArray[3] = (byte)cmd.FunctionType;
            packetArray[4] = (byte)(cmd.Address >> 8);
            packetArray[5] = (byte)(cmd.Address & 0xFF);
            packetArray[6] = (byte)(cmd.Quantity >> 8);
            packetArray[7] = (byte)(cmd.Quantity & 0xFF);

            int crc = CRC16.Calculate(packetArray, TX_PACKET_SIZE - 2);
            packetArray[8] = (byte)(crc & 0xFF);
            packetArray[9] = (byte)(crc >> 8);

            return packetArray;
        }

        /// <summary>預估回應的封包長度。</summary>
        /// <param name="funtionType">指令型態。</param>
        /// <param name="quantity">傳送內容數量</param>
        /// <returns></returns>
        public static int EstimateResponsePacketLength(FunctionType funtionType, byte quantity)
        {
            return (byte)funtionType < 5 ? (quantity * 2 + 7) : 10;
        }

        /// <summary>解析封包格式內容。</summary>
        /// <param name="recvDataArray">接收資料暫存器。</param>
        /// <param name="recvLength">接收資料長度。</param>
        /// <param name="cmd">回傳指令內容。</param>
        /// <returns>是否解析成功。</returns>
        public static bool Parse(byte[] recvDataArray, int recvLength, ref IAppCommand cmd)
        {
            return cmd.Parse(recvDataArray, recvLength);
        }

        /// <summary>確認回應的資料內容格式是否正確。</summary>
        /// <param name="deviceID">裝置位址。</param>
        /// <param name="sendingCmd">傳送應用程式碼指令。</param>
        /// <param name="recvDataArray">接收資料陣列。</param>
        /// <param name="recvLength">接收資料長度。</param>
        /// <returns>回應內容是否正確。</returns>
        /// 
        /// 目前只考慮 PSU 裝置，尚未考慮到 CSU/SBOX
        ///
        public static bool ValidateResponsePacket(byte deviceID, IAppCommand sendingCmd, byte[] recvDataArray, int recvLength)
        {
            bool isSuccess = false;
            byte functionCode = recvDataArray[RX_PACKET_FUNCTION_BYTE];
            int crc = CRC16.Calculate(recvDataArray, recvLength - 2); 

            if (functionCode == (byte)FunctionType.Read)
            {
                if (recvDataArray[RX_PACKET_HEAD_BYTE] == TX_PACKET_HEAD &&
                    recvDataArray[RX_PACKET_SOURCE_BYTE] == deviceID &&
                    recvDataArray[RX_PACKET_DESTINATION_BYTE] == TX_PACKET_MASTER &&
                    recvDataArray[RX_PACKET_FUNCTION_BYTE] == (byte)sendingCmd.FunctionType &&
                    recvDataArray[RX_PACKET_DATA_LENGTH_BYTE] == recvLength - 7 &&
                    recvDataArray[recvLength - 2] == (crc & 0xFF) &&
                    recvDataArray[recvLength - 1] == (crc >> 8))
                {
                    isSuccess = true;
                }
            }
            else if (functionCode == (byte)FunctionType.Write)
            {
                // ※ PSU 『寫』的時候，回應封包的來源跟目不互換。CSU、SBox皆會互換!!
                if (recvDataArray[RX_PACKET_HEAD_BYTE] == TX_PACKET_HEAD &&
                    recvDataArray[RX_PACKET_SOURCE_BYTE] == TX_PACKET_MASTER &&
                    recvDataArray[RX_PACKET_DESTINATION_BYTE] == deviceID &&
                    recvDataArray[RX_PACKET_FUNCTION_BYTE] == (byte)sendingCmd.FunctionType &&
                    recvDataArray[recvLength - 2] == (crc & 0xFF) &&
                    recvDataArray[recvLength - 1] == (crc >> 8))
                {
                    isSuccess = true;
                }
            }

            return isSuccess;
        }
    }
}
