﻿using System;
using System.Diagnostics;
using PH.Bootloader.RS485.Communication;
using PH.Bootloader.RS485.PIC32_2.Protocol.Command.AppCode;
using PH.Bootloader.RS485.PIC32_2.Protocol.Command.BootCode;

namespace PH.Bootloader.RS485.PIC32_2.Protocol
{
    /// <summary>實作 Bootloader 2.0 通訊協定之類別。</summary>
    class ProtocolHandler
    {
        private SerialPortHandler   comPort;
        private Stopwatch           recvStopwatch;

        /// <summary>取得封包接收錯誤計數。</summary>
        public long RxPacketErrorCount { get; internal set; }
        /// <summary>取得封包傳送計數。</summary>
        public long TxPacketCount { get; internal set; }

        /// <summary>初始化 ProtocolHandler 類別的新執行個體。</summary>
        public ProtocolHandler()
        {
            recvStopwatch = new Stopwatch();
        }

        /// <summary>指定『已開啟』之串列通訊埠。</summary>
        /// <param name="comPort">實體串列通訊埠。</param>
        /// <returns>串列通訊埠是否已開啟。</returns>
        public bool AssignComPort(SerialPortHandler comPort)
        {
            this.comPort = comPort;
            if (comPort == null) {
                throw new Exception("No specify communication for bootloader 2.0 protocol.");
            }

            return comPort.IsOpen;
        }

        /// <summary>執行應用程式碼指令。</summary>
        /// <param name="deviceID">裝置位址。</param>
        /// <param name="cmd">應用程式碼指令。</param>
        /// <returns>是否成功。</returns>
        public bool ExecuteAppCodeCommandNotResp(byte deviceID, IAppCommand cmd)
        {
            byte[] txBuffer = AppCodePacketFactory.Create(deviceID, cmd);
            Send(txBuffer, txBuffer.Length);

            return true;
        }

        /// <summary>執行應用程式碼指令。</summary>
        /// <param name="deviceID">裝置位址。</param>
        /// <param name="cmd">應用程式碼指令。</param>
        /// <param name="isNotResponse">不須回應。</param>
        /// <returns>是否成功。</returns>
        public bool ExecuteAppCodeCommand(byte deviceID, IAppCommand cmd)
        {
            bool isSuccess = false;

            byte[] txBuffer = AppCodePacketFactory.Create(deviceID, cmd);
            int estimatedLength = AppCodePacketFactory.EstimateResponsePacketLength(cmd.FunctionType, (byte)(cmd.Quantity & 0xFF));

            byte[] rxBuffer;
            bool isReceived = comPort.SendExpectedLengthReceive(txBuffer, txBuffer.Length, estimatedLength, out rxBuffer);
            if (isReceived)
            {
                if (AppCodePacketFactory.ValidateResponsePacket(deviceID, cmd, rxBuffer, rxBuffer.Length))
                {
                    isSuccess = AppCodePacketFactory.Parse(rxBuffer, rxBuffer.Length, ref cmd);
                }
            }

            CalculatePacketErrorCount(isSuccess);

            return isSuccess;
        }

        /// <summary>執行啟動程式碼指令。</summary>
        /// <param name="cmd">啟動程式碼指令。</param>
        /// <returns>是否成功。</returns>
        public bool ExecuteBootCodeCommand(IBootCommand cmd)
        {
            bool isSuccess = false;

            byte[] txBuffer = BootCodePacketFactory.Create(cmd);
            int estimatedLength = BootCodePacketFactory.EstimateResponsePacketLength(cmd);

            byte[] rxBuffer;
            bool isReceived = comPort.SendExpectedLengthReceive(txBuffer, txBuffer.Length, estimatedLength, out rxBuffer);
            if (isReceived) {
                isSuccess = BootCodePacketFactory.Parse(rxBuffer, rxBuffer.Length, ref cmd);
            }

            CalculatePacketErrorCount(isSuccess);

            return isSuccess;
        }

        /// <summary>在固定時間內，讀取一個位元組。</summary>
        /// <param name="data">讀取資料。</param>
        /// <param name="msec">讀取等待逾時時間。</param>
        /// <returns><是否讀到資料。/returns>
        public bool ReadByteAtCertainTime(out byte data, int msec)
        {
            recvStopwatch.Reset();
            recvStopwatch.Start();

            bool isGot = false;
            data = 0x00;
            while (true)
            {
                if (comPort.ReadByte(out data)) {
                    isGot = true;
                    break;
                }

                if (recvStopwatch.Elapsed.TotalMilliseconds >= msec) {
                    break;
                }
            }

            recvStopwatch.Stop();

            return isGot;
        }

        /// <summary>讀取一個位元組。</summary>
        /// <returns>資料位元組。</returns>
        public byte ReadByte()
        {
            return comPort.ReadByte();
        }

        /// <summary>資料傳送。</summary>
        /// <param name="txBuffer">傳送資料暫存器。</param>
        /// <param name="txLength">傳送資料長度。</param>
        public void Send(byte[] txBuffer, int txLength)
        {
            comPort.Send(txBuffer, txLength);
        }

        private void CalculatePacketErrorCount(bool isSuccess)
        {
            if (!isSuccess) {
                RxPacketErrorCount = RxPacketErrorCount >= long.MaxValue ? 0 : RxPacketErrorCount + 1;
            }
            TxPacketCount = TxPacketCount >= long.MaxValue ? 0 : TxPacketCount + 1;
        }
    }
}
