﻿using System;
using System.Text;

namespace PH.Bootloader.RS485.PIC32_2.Protocol
{
    /// <summary>更新用記憶體處理類別。</summary>
    class UpgradeMemoryHandler
    {
        private const int   UPGRADE_MEMORY_BYTE_MAX         = 1 * 1024 * 1024; // 1M Byte
        private const int   BOOTCODE2_MEMORY_BYTE_OFFSET    = UPGRADE_MEMORY_BYTE_MAX/2;

        private const int   BOOTLOADER1_FLASH_MIN           = 0x1D000000;
        private const int   BOOTLOADER1_FLASH_MAX           = 0x1D005FFF;
        private const int   BOOTLOADER2_FLASH_MIN           = 0x1FC00000;
        private const int   BOOTLOADER2_FLASH_MAX           = 0x1FC02FFF; // 0x1FC02FEF
        private const int   BOOTLOADER3_FLASH_MIN           = 0x1FC03000; // 0x1FC02FF0  
        private const int   BOOTLOADER3_FLASH_MAX           = 0x1FC03000;  
        private const int   APP_FLASH_MIN                   = 0x1D006000;
        private const int   APP_FLASH_MAX                   = 0x1D042FFF;
        private const int   BACKUP_FLASH_MIN                = 0x1D043000;
        private const int   BACKUP_FLASH_MAX                = 0x1D07FFFF;

        private const int   UPGRADE_BLOCK_SIZE              = 128;
        
        /// <summary>設定、取得記憶體索引最小值。</summary>
        public int          MemIndexMin { get; internal set; }
        /// <summary>設定、取得記憶體索引最大值。</summary>
        public int          MemIndexMax { get; internal set; }
        /// <summary>設定、取得記憶體總位元數。</summary>
        public int          TotalMemByte { get; internal set; }
        /// <summary>設定、取得讀取記憶體區塊大小。</summary>
        public int          MemBlockSize { get; internal set; }
        /// <summary>取得記憶體備妥旗標。</summary>
        public bool         IsMemReady { get { return TotalMemByte > 0; } }

        private ASCIIEncoding   asciiEncoding;
        private byte[]          memoryArray;
        private byte[]          signArray;
        private bool            isUseAppBackupZone;

        /// <summary>初始化 UpgradeMemoryHandler 類別的新執行個體。</summary>
        public UpgradeMemoryHandler()
        {
            asciiEncoding = new ASCIIEncoding();

            MemIndexMin = int.MaxValue;
            MemIndexMax = int.MinValue;
            memoryArray = new byte[UPGRADE_MEMORY_BYTE_MAX];

            signArray = new byte[4];

            MemBlockSize = UPGRADE_BLOCK_SIZE;
        }

        /// <summary>清除。</summary>
        public void Clear()
        {
            MemIndexMin = int.MaxValue;
            MemIndexMax = int.MinValue;
            TotalMemByte = 0;

            for (int i = 0; i< UPGRADE_MEMORY_BYTE_MAX; i++) {
                memoryArray[i] = 0xFF;
            }

            isUseAppBackupZone = false;
        }

        private int memIndex;

        /// <summary>填入記憶體。</summary>
        /// <param name="address">位址。</param>
        /// <param name="value">數值。</param>
        /// <param name="upgradeType">更新碼型態。</param>
        /// <returns></returns>
        public bool FillMemory(int address, byte value, UpgradeImageType upgradeType)
        {
            try
            {
                if (upgradeType == UpgradeImageType.AppCode ||
                    upgradeType == UpgradeImageType.AppCodeByBootCode)
                {
                    if (ValidAppCodeAddress(address)) 
                    {
                        // 0x1D006000 map to 0
                        memIndex = address - APP_FLASH_MIN;
                        memoryArray[memIndex] = value;
                        CalculateMemoryRange(memIndex);
                        return true;
                    }
                    else if (ValidBackupAddress(address))
                    {
                        // 2019/06/18 因應外掛 flash 的硬體配置，放寬成 Backup 區塊也可寫
                        memIndex = address - APP_FLASH_MIN;
                        memoryArray[memIndex] = value;
                        CalculateMemoryRange(memIndex);
                        isUseAppBackupZone = true;
                        return true;
                    }
                    else if (ValidBootCodeAddress2(address) ||
                             ValidBootCodeAddress3(address))
                    {
                        // 瑱賢: AppCode 會包含到一部份 BootCode ，但還不知道原因 → 就是這裡
                        // ValidBootCodeAddress2 * 44 筆
                        // ValidBootCodeAddress3 * 16 筆
                        return true;
                    }
                    else
                    {
                        Clear();
                    }
                }
                else if (upgradeType == UpgradeImageType.BootCode)
                {
                    // 更新 Boot Code 時，『一定不能』出現 App Code 的資料
                    if (ValidBootCodeAddress1(address) )
                    {
                        // 0x1D000000 map to 0
                        memIndex = address - BOOTLOADER1_FLASH_MIN;
                        memoryArray[memIndex] = value;
                        CalculateMemoryRange(memIndex);
                        return true;
                    }
                    else if (ValidBootCodeAddress2(address))
                    {
                        // 、0x1FC00000 map to 『BOOTCODE2_MEMORY_BYTE_OFFSET』
                        memIndex = address - BOOTLOADER2_FLASH_MIN + BOOTCODE2_MEMORY_BYTE_OFFSET;
                        memoryArray[memIndex] = value;
                        CalculateMemoryRange(memIndex);
                        return true;
                    }
                    else if (ValidBootCodeAddress3(address))
                    {
                        // 瑱賢: 不處理
                        return true;
                    }
                    else
                    {
                        Clear();
                    }
                }
            }
            catch (Exception)
            {
            }

            return false;
        }

        /// <summary>計算記憶體範圍。</summary>
        /// <param name="memIndex">記憶體索引。</param>
        private void CalculateMemoryRange(int memIndex)
        {
            if (memIndex <= MemIndexMin) {
                MemIndexMin = memIndex;
            }

            if (memIndex >= MemIndexMax) {
                MemIndexMax = memIndex;
            }

            TotalMemByte++;
        }

        /// <summary>轉換記憶體指標到記憶體位址。</summary>
        /// <param name="upgradeType">更新碼型態。</param>
        /// <param name="indexMem">記憶體位置索引。</param>
        /// <returns>對應的位址。</returns>
        public int ConvertMemoryIndexToAddress(UpgradeImageType upgradeType, int indexMem)
        {
            if (upgradeType == UpgradeImageType.AppCode ||
                upgradeType == UpgradeImageType.AppCodeByBootCode)
            {
                return indexMem + APP_FLASH_MIN;
            }
            else if (upgradeType == UpgradeImageType.BootCode)
            {
                if (indexMem < BOOTCODE2_MEMORY_BYTE_OFFSET)
                {
                    return indexMem + BOOTLOADER1_FLASH_MIN;
                }
                else
                {
                    return indexMem - BOOTCODE2_MEMORY_BYTE_OFFSET + BOOTLOADER2_FLASH_MIN;
                }
            }

            return -1;
        }

        /// <summary>取得區域索引。</summary>
        /// <param name="address"></param>
        /// <returns></returns>
        public byte GetZoneIndex(int address)
        {
            if (ValidBootCodeAddress2(address))
            {
                return 0x01;
            }
            else
            {
                return 0x00;
            }
        }

        /// <summary>依序(記憶體區塊)取得內容，直到結束。</summary>
        /// <param name="indexMem">記憶體位置索引。</param>
        /// <param name="blockArray">區塊資料陣列。</param>
        /// <returns>是否取得。</returns>
        public bool GetMemory(int indexMem, ref byte[] blockArray)
        {
            try
            {
                if (indexMem + UPGRADE_BLOCK_SIZE - 1 <= UPGRADE_MEMORY_BYTE_MAX)
                {
                    Array.Copy(memoryArray, indexMem, blockArray, 0, UPGRADE_BLOCK_SIZE);
                    return true;
                }
                
                return false;
            }
            catch (Exception)
            {
                return false;
            }
        }

        /// <summary>確認應用程式碼識別符號。</summary>
        /// <param name="devType">裝置型態。</param>
        /// <returns>是否符合。</returns>
        public bool VerifyAppCodeSign(DeviceType devType)
        {
            int indexSign = isUseAppBackupZone ?
                            0x1D07FFF0 - APP_FLASH_MIN :
                            0x1D042FF0 - APP_FLASH_MIN;

            signArray[0] = memoryArray[indexSign];
            signArray[1] = memoryArray[indexSign + 1];
            signArray[2] = memoryArray[indexSign + 2];
            signArray[3] = memoryArray[indexSign + 3];

            string strSign = asciiEncoding.GetString(signArray).TrimEnd();
            return string.Equals(strSign, devType.HexKey);
        }

        /// <summary>確認 應用程式碼 Check Sum。</summary>
        /// <returns>是否正確。</returns>
        public bool VerifyAppCodeCheckSum()
        {
            int indexCRC = isUseAppBackupZone ?
                            0x1D07FFFC - APP_FLASH_MIN :
                            0x1D042FFC - APP_FLASH_MIN;

            // little-endian (高位元放在高記憶體位址)
            uint crcMem = BitConverter.ToUInt32(memoryArray, indexCRC);

            // 相關細節請參考「MPLAB XC8 C Compiler User’s Guide」的 Chapter 8 Utilities。
            byte[] crcBytes = new byte[4];
            uint crcCal = 0;
            for (int i = 0; i< indexCRC; i+=4) {
                Array.Copy(memoryArray, i, crcBytes, 0, 4);
                crcCal += BitConverter.ToUInt32(crcBytes, 0); ;
            }

            return crcMem == crcCal;
        }

        /// <summary>確認啟動程式碼識別符號。</summary>
        /// <param name="devType"></param>
        /// <returns></returns>
        public bool VerifyBootCodeSign(DeviceType devType)
        {
            int indexSign = 0x1D005FF0 - BOOTLOADER1_FLASH_MIN;

            signArray[0] = memoryArray[indexSign];
            signArray[1] = memoryArray[indexSign + 1];
            signArray[2] = memoryArray[indexSign + 2];
            signArray[3] = memoryArray[indexSign + 3];

            string strSign = asciiEncoding.GetString(signArray).TrimEnd();
            return string.Equals(strSign, devType.HexKey);
        }

        /// <summary>確認啟動程式碼 Check Sum。</summary>
        /// <returns></returns>
        public bool VerifyBootCodeCheckSum()
        {
            int indexCRC = 0x1D005FFC - BOOTLOADER1_FLASH_MIN;

            // little-endian (高位元放在高記憶體位址)
            uint crcMem = BitConverter.ToUInt32(memoryArray, indexCRC);

            // 相關細節請參考「MPLAB XC8 C Compiler User’s Guide」的 Chapter 8 Utilities。
            byte[] crcBytes = new byte[4];
            uint crcCal = 0;
            for (int i = 0; i < indexCRC; i += 4)
            {
                Array.Copy(memoryArray, i, crcBytes, 0, 4);
                crcCal += BitConverter.ToUInt32(crcBytes, 0); ;
            }

            return crcMem == crcCal;
        }

        private bool ValidAppCodeAddress(int address)
        {
            return (address >= APP_FLASH_MIN && address <= APP_FLASH_MAX);
        }

        private bool ValidBackupAddress(int address)
        {
            return (address >= BACKUP_FLASH_MIN && address <= BACKUP_FLASH_MAX);
        }

        private bool ValidBootCodeAddress1(int address)
        {
            return (address >= BOOTLOADER1_FLASH_MIN && address <= BOOTLOADER1_FLASH_MAX);
        }

        private bool ValidBootCodeAddress2(int address)
        {
            return (address >= BOOTLOADER2_FLASH_MIN && address <= BOOTLOADER2_FLASH_MAX);
        }

        private bool ValidBootCodeAddress3(int address)
        {
            return (address >= BOOTLOADER3_FLASH_MIN && address <= BOOTLOADER3_FLASH_MAX);
        }
    }
}
