﻿using Net.Core.Communication.Serial.Sync;

namespace PH.Bootloader.RS485
{
    /// <summary>RS-485  啟動程序安裝工具類別。</summary>
    /// 2018/11/16 Alston不習慣獨體模式，再封裝一層!!
    public class RS485Bootloader
    {
        private RS485BootloaderFactory  bootloaderFactory;
        private IBootloader             bootloader;
        private ISyncCommunicationPlus  comPort;

        /// <summary>取得啟動程序名稱。</summary>
        public string   Name { get; private set; }
        /// <summary>取得程序執行中旗標。</summary>
        public bool     IsProcedureRunning { get { return bootloader.IsProcedureRunning; } }
        /// <summary>取得是否由 PIC 擔任代理燒錄的腳色 (透過 CAN Bus)(僅有 DSP 有該功能)。</summary>
        public bool     IsCanAgent { get { return bootloader.IsCanAgent; } set { bootloader.IsCanAgent = value; } }

        /// <summary>初始化 RS485Bootloader 實體物件。</summary>
        public RS485Bootloader()
        {
            bootloaderFactory = RS485BootloaderFactory.GetInstance();
            bootloader = null;
        }

        /// <summary>取消所有正在執行的程序。</summary>
        public void Abort()
        {
            if (bootloader != null && bootloader.IsProcedureRunning) {
                bootloader?.Abort();
            }
        }

        /// <summary>指定串列通訊埠。</summary>
        /// <param name="comPort">串列通訊埠介面。</param>
        /// <returns></returns>
        public bool AssignComPort(ISyncCommunicationPlus comPort)
        {
            this.comPort = comPort;

            if (bootloader != null) {
                return bootloader.AssignComPort(comPort);
            }
            return false;
        }

        /// <summary>檢查 App Code 是否存在。</summary>
        /// <param name="device"></param>
        /// <returns></returns>
        public bool CheckAppCode(DeviceInformation device)
        {
            if (bootloader != null) {
                return bootloader.CheckAppCode(device);
            }
            return false;
        }

        /// <summary>檢查裝置。</summary>
        /// <param name="device">目標裝置。</param>
        public bool CheckDevice(DeviceInformation device)
        {
            if (bootloader != null) {
                return bootloader.CheckDevice(device);
            }
            return false;
        }

        /// <summary>準備記憶體內容。</summary>
        /// <param name="device">目標裝置。</param>
        /// <param name="filePath">印象檔案路徑與名稱。</param>
        public void ParseHexToMemory(DeviceInformation device, string filePath)
        {
            bootloader.ParseHexToMemory(device, filePath);
        }

        /// <summary>設定啟動程序型態。</summary>
        /// <param name="deviceType"></param>
        public void SetBootloaderType(DeviceType deviceType)
        {
            if (bootloader != null)
            {
                bootloader.HexParsed -= Bootloader_HexParsed;
                bootloader.MessageShown -= Bootloader_MessageShown;
                bootloader.ProcedureCompleted -= Bootloader_ProcedureCompleted;
                bootloader.ProcedureProgressChanged -= Bootloader_ProcedureProgressChanged;
                bootloader.UploadInformationChanged -= Bootloader_UploadInformationChanged;
            }

            bootloader = bootloaderFactory.Create(deviceType);
            bootloader.AssignComPort(comPort);
            bootloader.HexParsed += Bootloader_HexParsed;
            bootloader.MessageShown += Bootloader_MessageShown;
            bootloader.ProcedureCompleted += Bootloader_ProcedureCompleted;
            bootloader.ProcedureProgressChanged += Bootloader_ProcedureProgressChanged;
            bootloader.UploadInformationChanged += Bootloader_UploadInformationChanged;
            Name = bootloader.Name;
        }

        /// <summary>上載。</summary>
        /// <param name="device">目標裝置。</param>
        /// <returns></returns>
        public bool Upload(DeviceInformation device)
        {
            if (bootloader != null) {
                return bootloader.Upload(device);
            }
            return false;
        }

        private void Bootloader_HexParsed(object sender, string[] records)
        {
            HexParsed?.Invoke(sender, records);
        }

        private void Bootloader_MessageShown(object sender, string message)
        {
            MessageShown?.Invoke(sender, message);
        }

        private void Bootloader_ProcedureCompleted(object sender, ProcedureType procType, BootloadResult result)
        {
            ProcedureCompleted?.Invoke(sender, procType, result);
        }

        private void Bootloader_ProcedureProgressChanged(object sender, ProcedureType procType, int progressPercentage)
        {
            ProcedureProgressChanged?.Invoke(sender, procType, progressPercentage);
        }

        private void Bootloader_UploadInformationChanged(object sender, long current, long total)
        {
            UploadInformationChanged?.Invoke(sender, current, total);
        }

        /// <summary>發生在解析出 Hex 內容後。</summary>
        public event HexParsedEventHandler                  HexParsed;
        /// <summary>發生在更新程序信息顯示時。</summary>
        public event MessageShownEventHandler               MessageShown;
        /// <summary>發生在程序執行完成時。</summary>
        public event ProcedureCompletedEventHandler         ProcedureCompleted;
        /// <summary>發生在程序進度改變時。</summary>
        public event ProcedureProgressChangedEventHandler   ProcedureProgressChanged;
        /// <summary>發生在資料更新改變時。</summary>
        public event UploadInformationChangedEventHandler   UploadInformationChanged;
    }
}
