﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Net.NetworkInformation;
using System.Text;
using FirmwareHeaderBuilder.Model;
using FirmwareHeaderBuilder.Model.ModelNaming;
using PhihongEv.Lib.Convert;

namespace FirmwareHeaderBuilder.Controller
{
    /// <summary>表示檔頭創建之處理類別。</summary>
    class HeaderBuilder
    {
        private string      imagePath;
        private byte[]      imageDataArray;
        private bool        isImageReady;
        private byte[]      modelNameArray;
        private bool        isModelNameReady;
        private ImageType   imageType;
        private CRC32       crc32Handler;

        private string      modelName;

        /// <summary>初始化 HeaderBuilder 類別的新執行個體。</summary>
        public HeaderBuilder()
        {
            imageDataArray = null;
            isImageReady = false;
            modelNameArray = null;
            isModelNameReady = false;

            modelName = "";

            crc32Handler = CRC32.GetInatance();
        }

        /// <summary>創建。</summary>
        /// <returns></returns>
        public ErrorCode Build()
        {
            if (!isImageReady) {
                return ErrorCode.ImageDataNotReady;
            }
            else if (!isModelNameReady)
            {
                return ErrorCode.ModelNameNotReady;
            }

            byte[] dataSegment1 = new byte[16];
            byte[] dataSegment2 = new byte[16];
            byte[] dataSegment3 = new byte[16];

            Array.Clear(dataSegment1, 0, 16);
            for (int i = 0; i < 16; i++) {
                dataSegment2[i] = 0xFF;
                dataSegment3[i] = 0xFF;
            }

            try
            {
                // 0000h
                Array.Copy(modelNameArray, 0, dataSegment1, 0, modelNameArray.Length);
                // 0001h
                byte[] imgLenBytes = GetImgLenBytes(imageDataArray);
                byte[] timeBytes = GetCreateTimeBytes();
                Array.Copy(imageType.CodeArray, 0, dataSegment2, 0, 4);
                Array.Copy(imgLenBytes, 0, dataSegment2, 4, 4);
                Array.Copy(timeBytes, 0, dataSegment2, 8, 8);
                // 0002h
                Array.Copy(timeBytes, 8, dataSegment3, 0, 2);
                byte[] crcBytes = GetCRC32(dataSegment1, dataSegment2, dataSegment3, imageDataArray);
                Array.Copy(crcBytes, 0, dataSegment3, 2, 4);
            }
            catch
            {
                return ErrorCode.MemoryAllocationError;
            }

            List<byte> contents = new List<byte>();
            contents.AddRange(dataSegment1);
            contents.AddRange(dataSegment2);
            contents.AddRange(dataSegment3);
            contents.AddRange(imageDataArray);

            try
            {
                string fullPath = Path.GetFullPath(imagePath);
                string ext = Path.GetExtension(imagePath);
                string newFilePath;

                if (string.IsNullOrEmpty(ext)) {
                    newFilePath = fullPath + "_" + modelName;
                } else {
                    newFilePath = fullPath.Replace(ext, "_" + modelName + ext);
                }

                File.WriteAllBytes(newFilePath, contents.ToArray());
            }
            catch
            {
                return ErrorCode.BuiltFileError;
            }

            return ErrorCode.BuiltSuccess;
        }

        /// <summary>檢查檔頭是否存在。</summary>
        /// <param name="filePath"></param>
        public bool CheckHeaderExist(string filePath)
        {
            bool isExist = false;

            if (!string.IsNullOrEmpty(filePath))
            {
                if (File.Exists(filePath))
                {
                    try
                    {
                        byte[] imgDataBytes = File.ReadAllBytes(filePath);

                        if (imgDataBytes != null && imgDataBytes.Length > 48)
                        {
                            string mainType = Encoding.ASCII.GetString(imgDataBytes, 0, 2);
                            string inputConfig = Encoding.ASCII.GetString(imgDataBytes, 2, 1);
                            string saftyRegulation = Encoding.ASCII.GetString(imgDataBytes, 3, 1);

                            if (mainType.ToMajorType() != null &&
                                inputConfig.ToInputConfig() != null &&
                                saftyRegulation.ToSafetyRegulation() != null)
                            {
                                byte[] segment1 = new byte[16];
                                byte[] segment2 = new byte[16];
                                byte[] segment3 = new byte[16];
                                byte[] imageArray = new byte[imgDataBytes.Length - 48];

                                Array.Copy(imgDataBytes, 0, segment1, 0, 16);
                                Array.Copy(imgDataBytes, 16, segment2, 0, 16);
                                Array.Copy(imgDataBytes, 32, segment3, 0, 16);
                                Array.Copy(imgDataBytes, 48, imageArray, 0, imgDataBytes.Length - 48);

                                byte[] crc32 = GetCRC32(segment1, segment2, segment3, imageArray);

                                isExist = (crc32[0] == segment3[2] &&
                                           crc32[1] == segment3[3] &&
                                           crc32[2] == segment3[4] &&
                                           crc32[3] == segment3[5]);
                            }
                        }
                    }
                    catch
                    {
                    }
                }
            }

            return isExist;
        }

        /// <summary>清除映像檔相關資訊。</summary>
        public void ClearImage()
        {
            imagePath = string.Empty;
            isImageReady = false;
            imageDataArray = null;
        }

        private byte[] GetImgLenBytes(byte[] imageArray)
        {
            // Big Endian
            byte[] imgLenBytes = BitConverter.GetBytes(imageArray.Length);
            Array.Reverse(imgLenBytes);

            return imgLenBytes;
        }

        private byte[] GetCreateTimeBytes()
        {
            string strTime = DateTime.Now.ToString("yyyyMMddHH");
            return Encoding.ASCII.GetBytes(strTime);
        }

        private byte[] GetCRC32(byte[] segment1, byte[] segment2, byte[] segment3, byte[] imageArray)
        {
            List<byte> crcData = new List<byte>();
            crcData.AddRange(segment1);
            crcData.AddRange(segment2);
            crcData.AddRange(new byte[] { segment3[0], segment3[1] });
            crcData.AddRange(imageArray);

            ulong crc32 = crc32Handler.Calculate(crcData.ToArray(), crcData.Count);

            // Big Endian
            byte[] outcome = new byte[4];
            outcome[0] = (byte)((crc32 & 0xFF000000) >> 24);
            outcome[1] = (byte)((crc32 & 0x00FF0000) >> 16);
            outcome[2] = (byte)((crc32 & 0x0000FF00) >> 8);
            outcome[3] = (byte)(crc32 & 0x000000FF);

            return outcome;
        }

        /// <summary>匯入映像檔。</summary>
        /// <param name="filePath"></param>
        public ErrorCode ImportImage(string filePath)
        {
            ErrorCode errCode = ErrorCode.NoError;
            ClearImage();

            if (!string.IsNullOrEmpty(filePath))
            {
                if (File.Exists(filePath))
                {
                    try
                    {
                        imageDataArray = File.ReadAllBytes(filePath);
                        imagePath = filePath;
                        isImageReady = true;
                    }
                    catch
                    {
                        errCode = ErrorCode.ReadFileError;
                    }
                }
                else
                {
                    errCode = ErrorCode.FileNotExist;
                }
            }
            else
            {
                errCode = ErrorCode.FilePathEmpty;
            }

            return errCode;
        }

        /// <summary>匯入映像檔。</summary>
        /// <param name="filePath"></param>
        /// <param name="isHeaderExist"></param>
        public ErrorCode ImportImage(string filePath, bool isHeaderExist)
        {
            if (isHeaderExist)
            {
                ErrorCode errCode = ErrorCode.NoError;
                imagePath = string.Empty;
                isImageReady = false;
                imageDataArray = null;

                if (!string.IsNullOrEmpty(filePath))
                {
                    if (File.Exists(filePath))
                    {
                        try
                        {
                            byte[] tempDataArray = File.ReadAllBytes(filePath);
                            if (tempDataArray != null && tempDataArray.Length > 48)
                            {
                                imageDataArray = new byte[tempDataArray.Length - 48];
                                Array.Copy(tempDataArray, 48, imageDataArray, 0, imageDataArray.Length);
                                imagePath = filePath;
                                isImageReady = true;
                            }
                            else
                            {
                                errCode = ErrorCode.ReadFileError;
                            }
                        }
                        catch
                        {
                            errCode = ErrorCode.ReadFileError;
                        }
                    }
                    else
                    {
                        errCode = ErrorCode.FileNotExist;
                    }
                }
                else
                {
                    errCode = ErrorCode.FilePathEmpty;
                }

                return errCode;
            }
            else
            {
                return ImportImage(filePath);
            }
        }

        /// <summary>輸入產品型號。</summary>
        /// <param name="name"></param>
        /// <returns></returns>
        public ErrorCode InputModelName(string name)
        {
            ErrorCode errCode = ErrorCode.NoError;
            modelNameArray = null;
            isModelNameReady = false;

            if (!string.IsNullOrEmpty(name))
            {
                if (name.Length == 14)
                {
                    try
                    {
                        modelNameArray = Encoding.ASCII.GetBytes(name);
                        modelName = name;
                        isModelNameReady = true;
                    }
                    catch
                    {
                        errCode = ErrorCode.ModelNameInputTypeError;
                    }
                }
                else
                {
                    errCode = ErrorCode.ModelNameLenError;
                }
            }
            else
            {
                errCode = ErrorCode.ModelNameEmpty;
            }

            return errCode;
        }

        /// <summary>選擇映像檔型態。</summary>
        /// <param name="imaTypeIndex"></param>
        /// <returns></returns>
        public ErrorCode SelectImageType(ushort imaTypeIndex)
        {
            imageType = ImageType.Parse(imaTypeIndex);

            return imageType == null ? ErrorCode.ImageTypeIndexNotExist : ErrorCode.NoError;
        }
    }
}
