﻿using System;
using System.Drawing;
using System.IO;
using System.IO.Ports;
using System.Windows.Forms;
using Net.Core.Communication.Serial.Sync;
using PH.Bootloader.RS485;

namespace IntegratedBootloader
{
    public partial class Form1
    {
        private void ComOption_SelectedIndexChanged(object sender, EventArgs e)
        {
            string name = (string)ComOption.SelectedItem;

            if (name.Equals("OFF")) {
                if (syncComm.IsOpen) {
                    Billboard.AppendText("\r\nClosen communication port...\r\n");
                }
                syncComm.Close();
            } else {
                Billboard.AppendText("\r\nOpen communication port...\r\n");
                syncComm.Open(name, 115200);
            }
        }

        private void ComOption_DropDown(object sender, EventArgs e)
        {
            ComOption.Items.Clear();
            ComOption.Items.Add("OFF");
            ComOption.Items.AddRange(SerialPort.GetPortNames());
            ComOption.SelectedIndex = 0;
        }


        private void SyncComm_StatusChanged(SyncCommStatus status)
        {
            isCommReady = status == SyncCommStatus.Opened;

            if (isCommReady)
            {
                bootloader.AssignComPort(syncComm);
                bootloader.SetBootloaderType(deviceType);
            }

            Billboard.AppendText(string.Format("Communication port has been {0}.\r\n", isCommReady ? "opend" : "closed"));
            CommState.BackColor = isCommReady ? Color.Green : Color.Silver;
            if (isCommReady) {
                ControlIdleUI();
            } else {
                ControlAllUI(false);
            }
        }

        private void DeviceTypeOtion_SelectedIndexChanged(object sender, EventArgs e)
        {
            deviceType = DeviceType.ParseByIndex(DeviceTypeOtion.SelectedIndex);

            ControlIdleUI();

            if (isCommReady) {
                bootloader.SetBootloaderType(deviceType);

                HexFileText.Text = string.Empty;

                string devTypeName = deviceType.Name;
                Billboard.AppendText(string.Format("\r\n------------------------------------------------------------------------------------------------------------------ {0}↓\r\n", devTypeName));
            }

            string[] deviceNames = DeviceInformation.GetNameList(deviceType);
            DeviceNameOption.Items.Clear();
            DeviceNameOption.Items.AddRange(deviceNames);
            DeviceNameOption.SelectedIndex = 0;

            CanAgentChecker.Visible = deviceType.Equals(DeviceType.DSP_PSU) || deviceType.Equals(DeviceType.DSP_PFC);
            AppCodeBTN.Visible = deviceType.Equals(DeviceType.PIC_PSU) || deviceType.Equals(DeviceType.PIC_PFC);
        }

        private void DeviceNameOption_SelectedIndexChanged(object sender, EventArgs e)
        {
            string devName = (string)DeviceNameOption.SelectedItem;
            targetDevice = DeviceInformation.ParseByName(devName);
        }

        private void Button_Click(object sender, EventArgs e)
        {
            if (!isCommReady) {
                return;
            }

            Button button = sender as Button;
            string btnName = button.Name;

            if (btnName.Equals("CheckBTN"))
            {
                Billboard.AppendText("\r\n" + targetDevice.Name + " >>>>>>\r\n ");
                ControlRunningUI();
                bootloader.CheckDevice(targetDevice);
            }
            else if (btnName.Equals("BrowserBTN"))
            {
                OpenFileDialog openFileDialog = new OpenFileDialog();
                openFileDialog.Filter = "hex files (*.hex*)|*.hex*|All files (*.*)|*.*";
                openFileDialog.FilterIndex = 1;
                openFileDialog.RestoreDirectory = true;

                if (openFileDialog.ShowDialog() == DialogResult.OK) {
                    HexFileText.Text = openFileDialog.FileName;
                    bootloader.ParseHexToMemory(targetDevice, openFileDialog.FileName);
                }
            }
            else if (btnName.Equals("UploadBTN"))
            {
                Billboard.AppendText("\r\n" + targetDevice.Name + " >>>>>>\r\n ");
                ControlRunningUI();

                bootloader.IsCanAgent = CanAgentChecker.Visible && CanAgentChecker.Checked;
                bootloader.Upload(targetDevice);
            }
            else if (btnName.Equals("AbortBTN"))
            {
                bootloader.Abort();
            }
            else if (btnName.Equals("AppCodeBTN"))
            {
                Billboard.AppendText("\r\nCheck application code...\r\n");
                bool isReady = bootloader.CheckAppCode(targetDevice);
                Billboard.AppendText(isReady ? "Application code is ready\r\n" :
                                               "Application code is not ready!!\r\n");
            } 
        }

        private void HexFileText_DragEnter(object sender, DragEventArgs e)
        {
            if (!isCommReady) {
                return;
            }

            e.Effect = e.Data.GetDataPresent(DataFormats.FileDrop) ? DragDropEffects.All : DragDropEffects.None;
        }

        private void HexFileText_DragDrop(object sender, DragEventArgs e)
        {
            if (!isCommReady) {
                return;
            }

            HexFileText.Text = ((string[])e.Data.GetData(DataFormats.FileDrop)).GetValue(0).ToString();

            if (File.Exists(HexFileText.Text))
            {
                if (ValidFileExtension(HexFileText.Text))
                {
                    bootloader.ParseHexToMemory(targetDevice, HexFileText.Text);
                }
                else
                {
                    HexFileText.Text = string.Empty;
                    Billboard.AppendText("\r\nParse image file...\r\n");
                    Billboard.AppendText("           Err: Image filename extension must be *.hex!!\r\n");
                }
            }
            else
            {
                Billboard.AppendText("\r\nParse image file...\r\n");
                Billboard.AppendText("           Err: Image file dose not exist!!\r\n"); 
            }
        }

        private bool ValidFileExtension(string filePath)
        {
            string ext = Path.GetExtension(filePath);
            return ext.Equals(".hex");
        }

        private void Bootloader_ProcedureProgressChanged(object sender, ProcedureType procType, int progressPercentage)
        {
            UpdateProgress.Value = progressPercentage;
            PercentageText.Text = string.Format("{0}%", progressPercentage);
        }

        private void Bootloader_UploadInformationChanged(object sender, long current, long total)
        {
            UploadInfoText.Text = string.Format("{0}/{1}", current, total);
        }

        private void Bootloader_ProcedureCompleted(object sender, ProcedureType procType, BootloadResult result)
        {
            if (result == BootloadResult.SerialPortReady)
            {

            }
            else if (result == BootloadResult.SerialPortNotReady)
            {

            }
            else if (result == BootloadResult.DetectDevice)
            {
                if (procType == ProcedureType.UploadCheckDevice)
                {
                    ControlRunningUI();
                }
                else
                {
                    // 臨時加上的
                    if (!string.IsNullOrEmpty(HexFileText.Text))
                    {
                        ControlImageReadyUI();
                    }
                    else
                    {
                        ControlIdleUI();
                    }
                }
            }
            else if (result == BootloadResult.DetectNothing ||
                     result == BootloadResult.DetectFailed)
            {
                ControlIdleUI();
            }
            
            else if (result == BootloadResult.FlashErased)
            {
                ControlRunningUI();
            }
            else if (result == BootloadResult.ImageFileReady)
            {
                ControlImageReadyUI();
            }
            else if (result == BootloadResult.MemoryAllocationReady)
            {

            }
            else if (
                     result == BootloadResult.MemoryAllocationFailed ||
                     result == BootloadResult.ImageFileNotExist ||
                     result == BootloadResult.ImageFileNotAppCode ||
                     result == BootloadResult.ImageFileNotMatch)
            {
                HexFileText.Text = string.Empty;
                ControlIdleUI();
            }
            else if (result == BootloadResult.EraseFlashFailed ||
                     result == BootloadResult.Uploaded ||
                     result == BootloadResult.UploadFailed)
            {
                ControlImageReadyUI();
            }
        }

        private void Bootloader_MessageShown(object sender, string message)
        {
            Billboard.AppendText(message);
        }
    }
}
