﻿using System;
using System.Drawing;
using System.IO;
using System.IO.Ports;
using System.Linq;
using System.Windows.Forms;
using System.Security.Principal;
using System.Reflection;
using Net.Core.Communication.Serial.Sync;
using PH.Bootloader.RS485;
using PH.Bootloader.RS485.PIC32;
using PH.Bootloader.RS485.Communication;

namespace PIC32Bootloader
{
    public partial class Form1 : Form
    {
        #region Fields
        private ISyncCommunicationPlus syncComm;
        private PicBootloader       bootloader;
        private string              filePath;
        private DeviceInformation   targetDevice;
        #endregion

        #region Constructor
        /// <summary>建構子。</summary>
        public Form1()
        {
            InitializeComponent();

            CheckForIllegalCrossThreadCalls = false;
            Text = Text + string.Format(" {0}{1}", "V", Assembly.GetEntryAssembly().GetName().Version);
            Initialize();
        }
        #endregion

        #region Methods : System
        private void Form1_Load(object sender, EventArgs e)
        {
            string strUserName = WindowsIdentity.GetCurrent().Name;
            string[] aryUserInfo = strUserName.Split(new string[] { "\\" }, StringSplitOptions.RemoveEmptyEntries);
            if (aryUserInfo.Length >= 2)
            {
                Billboard.AppendText(string.Format("Welcome~ {0}\r\n", aryUserInfo[1]));
            }
            else
            {
                Billboard.AppendText(string.Format("Welcome~ {0}\r\n", aryUserInfo[0]));
            }
        }

        private void CheckDevice_Click(object sender, EventArgs e)
        {
            bootloader.CheckDevice(targetDevice);
        }

        private void ImageBrowse_Click(object sender, EventArgs e)
        {
            if (openFileDialog1.ShowDialog() == DialogResult.OK)
            {
                filePath = openFileDialog1.FileName;
                FilePath.Text = Path.GetFileName(filePath);
                bootloader.ParseHexToMemory(targetDevice, filePath);
            }
        }

        private void WriteDevice_Click(object sender, EventArgs e)
        {
            if (File.Exists(filePath))
            {
                bootloader.Upload(targetDevice);
            }
            else
            {
                MessageBox.Show("Never select image file.", "Open file", MessageBoxButtons.OK, MessageBoxIcon.Warning);
            }
        }

        private void Abort_Click(object sender, EventArgs e)
        {
            bootloader.Abort();
        }

        private void AppCode_Click(object sender, EventArgs e)
        {
            bool isOK = bootloader.CheckAppCode(targetDevice);
            if (isOK)
            {
                Billboard.AppendText("Application code is ready\r\n");
            }
            else
            {
                Billboard.AppendText("Application code is not ready!!\r\n");
            }
        }
        #endregion

        #region Methods : Private
        private void Initialize()
        {
            syncComm = new SerialPortHandler();
            syncComm.StatusChanged += SyncSerialPort_StatusChanged; ;
            syncComm.SetSyncTime(50);

            bootloader = new PicBootloader();
            bootloader.ProcedureProgressChanged += Bootloader_ProcedureProgressChanged;
            bootloader.ProcedureCompleted += Bootloader_ProcedureCompleted;
            bootloader.MessageShown += Bootloader_MessageShown;

            ComOption.SelectedIndexChanged += ComOption_SelectedIndexChanged;
            ComOption.Items.Clear();
            ComOption.Items.Add("OFF");
            ComOption.Items.AddRange(SerialPort.GetPortNames());
            ComOption.SelectedIndex = 0;

            DeviceOption.SelectedIndexChanged += DeviceOption_SelectedIndexChanged;
            DeviceOption.Items.Clear();
            DeviceOption.Items.AddRange(DeviceInformation.ListDeviceName.ToList().ToArray());
            DeviceOption.SelectedIndex = 0;
        }

        private void ControlUI(bool enable)
        {
            DeviceOption.Enabled = enable;
            CheckDevice.Enabled = enable;
            ImageBrowse.Enabled = enable;
            WriteDevice.Enabled = enable;
            Abort.Enabled = enable;
            AppCode.Enabled = enable;
        }
        #endregion

        #region Methods : Callback
        private void ComOption_SelectedIndexChanged(object sender, EventArgs e)
        {
            string name = (string)ComOption.SelectedItem;
            if (name.Equals("OFF"))
            {
                syncComm.Close();
            }
            else
            {
                syncComm.Open(name, 115200);
                bootloader.AssignComPort(syncComm);
            }
        }

        private void DeviceOption_SelectedIndexChanged(object sender, EventArgs e)
        {
            targetDevice = DeviceInformation.ParseByIndex(DeviceOption.SelectedIndex);
        }

        private void SyncSerialPort_StatusChanged(SyncCommStatus status)
        {
            bool isOpen = status == SyncCommStatus.Opened;
            ComState.BackColor = isOpen ? Color.Green : Color.Silver;
            ControlUI(isOpen);
        }

        private void Bootloader_ProcedureProgressChanged(object sender, ProcedureType procType, int progressPercentage)
        {
            ProgressBar.Value = progressPercentage;
        }

        private void Bootloader_ProcedureCompleted(object sender, ProcedureType procType, BootloadResult result)
        {
            if (result == BootloadResult.EraseFlashFailed)
            {
                
            }
            else if (result == BootloadResult.ImageFileNotMatch)
            {
                
            }
            else if (result == BootloadResult.MemoryAllocationFailed)
            {
                
            }
            else if (result == BootloadResult.SerialPortNotReady)
            {
                
            }
            else if (result == BootloadResult.UploadFailed)
            {
                
            }
            else if (result == BootloadResult.DetectNothing)
            {
                
            }
            else if (result == BootloadResult.ImageFileNotExist)
            {
                
            }
            else if (result == BootloadResult.ImageFileNotAppCode)
            {
                
            }
        }

        private void Bootloader_MessageShown(object sender, string message)
        {
            Billboard.AppendText(message);
        }
        #endregion 
    }
}
